import os

from bamboo.common.command import CommandManager
from bamboo.wrappers.program_selector import allowed_refiner_args
from bamboo.wrappers.pdbutils import remove_modres_records

class RefinerObject(object):

    def __init__(self, time=True, verbose=True):

        self.allowedArgs = allowed_refiner_args
        # Settings
        self.time = time
        self.verbose = verbose
        self.runtime = -1.0
        self.timeout = 1800
        # Custom Init
        self._custom_init()

    def run_refinement(self, inpdb, inmtz, outdir, outfile, incif=None, flags=[]):
        """Run Refinement of a Structure"""

        # Prepare standard settings
        self._prepare_to_refine_structure(inpdb, inmtz, outdir, outfile, incif, flags)
        # Check to see if already run, and if not, run the custom part of the object - different for each program
        if os.path.exists(self.outpdb) and os.path.exists(self.outmtz) and os.path.exists(self.outlog):
            if self.verbose:
                print('\tRefinement already done - DOING NOTHING')
        else:
            if self.verbose:
                print('\tRefining using {!s}.'.format(self.name))
            cmd_line_args, std_inpt_args = self._create_program_arguments(inpdb, inmtz, outdir, outfile, incif, flags)
            # Initialise CommandManager
            self.Refiner = CommandManager(self.program)
            # Set Command-Line Args
            self.Refiner.SetArguments(cmd_line_args)
            # Set Standard Input
            self.Refiner.SetInput(std_inpt_args)
            # Set Parameters
            self.Refiner.SetParameters(timeout=self.timeout)
            # RUN
            self.Refiner.Run()
            # Calculate runtime (seconds)
            self.runtime = self.Refiner.runtime

            if self.verbose:
                print('\tFinished refinement - post-processing structure.')

            try:
                # General post-process
                self._standard_post_process(inpdb, inmtz, outdir, outfile, incif, flags)

                # Could add autoprocessing here...
                self._post_process(inpdb, inmtz, outdir, outfile, incif, flags)

            finally:
                self.write_log_file()

        return self.outpdb, self.outmtz, self.outlog

    def _prepare_to_refine_structure(self, inpdb, inmtz, outdir, outfile, incif, flags):
        """Set up the generic file names"""

        # Process outputfile
        if '/' in outfile:
            raise ValueError('outfile must be a file, not a path')
        # Record Template and Outdir
        self.outtemplate = os.path.join(outdir,outfile)
        self.outdir = outdir
        # Record Filenames
        self.outpdb = self.outtemplate+'.pdb'
        self.outmtz = self.outtemplate+'.mtz'
        self.outcif = self.outtemplate+'.cif'
        self.outlog = self.outtemplate+'.log'

    def _standard_post_process(self, inpdb, inmtz, outdir, outfile, incif, flags):
        """Perform standard operations on the output files of the refinement"""

        if os.path.exists(self.outpdb):
            remove_modres_records(self.outpdb)

        return

    def write_log_file(self):
        """Write the log file"""

        if self.verbose:
            print('\tWriting refinement logfile.')

        with open(self.outlog,'w') as logfile:
            # Write out the input command
            logfile.write('\nCOMMAND\n\n')
            logfile.write('\n'.join(self.Refiner.command))
            logfile.write('\nINPUT\n\n')
            logfile.write(self.Refiner.inp)
            # Write out & err
            logfile.write('\nSTDOUT\n\n')
            logfile.write(self.Refiner.out)
            logfile.write('\nSTDERR\n\n')
            logfile.write(self.Refiner.err)



import numpy
from scitbx.array_family import flex

def pairwise_dists(xyz1, xyz2):
    """Calculate pairwise distances between xyz1 and xyz2"""
    return numpy.array([flex.sqrt((xyz2 - c1).dot()) for c1 in xyz1])

def is_within(dist, coords_1, coords_2):
    """Checks if any of coords_1 is within dist of coords_2"""

    dist_sq = dist**2
    if len(coords_2) < len(coords_1):
        i1 = coords_2; i2 = coords_1
    else:
        i1 = coords_1; i2 = coords_2

    for c1 in i1:
        diffs = i2 - c1
        min_d_sq = flex.min(diffs.dot())
        if min_d_sq < dist_sq:
            return True
    return False



from scitbx.array_family import flex

def apply_symmetry_to_miller_array(chg_miller, ref_miller):
    """Applies the crystal symmetry, spacegroup info, and unit cell from ref_miller to change_miller - returns a miller array"""

    new_miller_set = chg_miller.set(crystal_symmetry=ref_miller.crystal_symmetry(), unit_cell=ref_miller.unit_cell(), space_group_info=ref_miller.space_group_info())
    new_miller_array = new_miller_set.array(data=chg_miller.data(), sigmas=chg_miller.sigmas())

    assert new_miller_array.is_similar_symmetry(ref_miller), 'Failure in applying symmetry from reference miller array'

    return new_miller_array

def check_miller_arrays_compatible_for_scaling(miller1, miller2, relative_length_tolerance=0.1, absolute_angle_tolerance=5):
    """Takes miller arrays and checks that they can be scaled together"""

    if not miller1.is_similar_symmetry(miller2, relative_length_tolerance=relative_length_tolerance, absolute_angle_tolerance=absolute_angle_tolerance):
        return False

    return True

def scale_miller_array_to_reference(chg_arr, ref_arr, by_bins=True, binning=None, n_bins=10):
    """Takes miller array, and scales it to another miller array"""

    if by_bins:
        # Apply or setup bins
        if binning:
            ref_binner = ref_arr.use_binning(binning)
            chg_binner = chg_arr.use_binning(binning)
        else:
            ref_binner = ref_arr.setup_binner(n_bins=n_bins)
            chg_binner = chg_arr.use_binning_of(ref_arr)

        # Create a blank copy of the chg_arr to be filled with the scaled values
        scl_arr = chg_arr.array(data=flex.double(chg_arr.size()), sigmas=flex.double(chg_arr.size()))

        # Scale!
        for i in ref_binner.range_used():
            # Select the indices in the bin
            ref_sel_bin = ref_binner.selection(i)
            chg_sel_bin = chg_binner.selection(i)

            # Select the reflections in the bin
            ref_arr_bin = ref_arr.select(ref_sel_bin)
            chg_arr_bin = chg_arr.select(chg_sel_bin)

            # Print raw ratio or means
#            print('<ref bin>/<scl bin> = {!s}'.format(ref_arr_bin.mean()/chg_arr_bin.mean()))

            # Scale the data in the bin - note it is the argument of 'scale' that gets scaled
            chg_arr_bin_scl = ref_arr_bin.scale(chg_arr_bin)

            # Add this scaled data to the new array
            scl_arr.data().set_selected(chg_sel_bin, chg_arr_bin_scl.data())

    else:
        # Simple scaling
        scl_arr = ref_arr.scale(chg_arr)

    scl_binner = scl_arr.use_binning_of(ref_arr)

    return scl_arr

def scale_amplitude_arrays_via_intensities(chg_arr_a, ref_arr_a, n_bins=20):
    """Take two amplitude arrays, converts them to intensities, scale them, then convert back to amplitudes"""

    assert ref_arr_a.is_similar_symmetry(chg_arr_a)

    # Convert both to intensity arrays for scaling
    ref_arr_i = ref_arr_a.as_intensity_array()
    ref_arr_i.set_observation_type_xray_intensity()
    chg_arr_i = chg_arr_a.as_intensity_array()
    chg_arr_i.set_observation_type_xray_intensity()

    # Scale mill2 to mill1
    chg_arr_i_scale = scale_miller_array_to_reference(chg_arr=chg_arr_i, ref_arr=ref_arr_i, n_bins=n_bins)
    chg_arr_i_scale.set_observation_type_xray_intensity()

    # Convert back to amplitudes
    chg_arr_a_scale = chg_arr_i_scale.as_amplitude_array()
    chg_arr_a_scale.set_observation_type_xray_amplitude()

    assert chg_arr_a_scale.is_similar_symmetry(chg_arr_a)

#    print('Correlation between unscaled and scaled amplitudes: {!s}'.format(round(chg_arr_a.correlation(chg_arr_a_scale).coefficient(),3)))

    return chg_arr_a_scale











import math
from rdkit import Chem

from bamboo.rdkit_utils.bonds.rotate import identify_rotatable_bond_atom_pairs

def calculate_dihedral_angle_differences(mol1, mol2):
    """Calculates the dihedral angle differences between rotatable bonds of mol1 and mol2"""

    # Get the dihedrals to calculate for both of the molecules (possibly multiple ways of overlaying the mols if symmetry exists)
    atom_sets = calculate_dihedral_atom_equivalences(mol1, mol2)
    # list of possible rmsds for the molecule
    differences = []
    # Iterate through and calculate the rmsd for each set of atom equivalences
    for mol1_atom_set, mol2_atom_set in atom_sets:
        # Calculate the dihedrals of both
        mol1_dihedrals = calculate_dihedral_angles(mol1, mol1_atom_set)
        mol2_dihedrals = calculate_dihedral_angles(mol2, mol2_atom_set)
        # Calculate the differences squared for each angle difference
        diffs = [an1-an2 for an1, an2 in zip(mol1_dihedrals,mol2_dihedrals)]
        # Append list of angle differences
        differences.append(diffs)

    return atom_sets, differences

def calculate_dihedral_atom_equivalences(mol1, mol2):
    """Gets the list that are paired in mol1 and mol2"""

    # Check that the mols are identical-ish
    if mol1.GetNumHeavyAtoms() != mol2.GetNumHeavyAtoms():
        raise EqualityError('Molecules are not identical (Num Atoms) {!s} != {!s}.\n{!s}\n{!s}'.format(mol1.GetNumHeavyAtoms(),mol2.GetNumHeavyAtoms(),Chem.MolToSmiles(mol1),Chem.MolToSmiles(mol2)))
    if mol1.GetNumBonds() != mol2.GetNumBonds():
        raise EqualityError('Molecules are not identical (Num Bonds) {!s} != {!s}:\n{!s}\n{!s}'.format(mol1.GetNumBonds(),mol2.GetNumBonds(),Chem.MolToSmiles(mol1), Chem.MolToSmiles(mol2)))

    # Gets a list of lists of atoms in mol1 (12,16,3, ...) that match the atoms in mol2 (1,2,3, ...)
    match_patterns = mol1.GetSubstructMatches(mol2, uniquify=False)
    # Get the quadruplets to calculate the dihedrals from for mol1
    mol1_atom_sets = identify_rotatable_bond_atom_pairs(mol1)
    num_atms = mol1.GetNumHeavyAtoms()
    # List for returning
    paired_atom_sets = []
    # Iterate through the different ways of overlaying the molecule (ensures we get the minimum rmsd)
    for match_pattern in match_patterns:
        # Translate from the atoms in mol1 to the atoms in mol2 (for this match_pattern)
        trans_dict = dict(zip(match_pattern, range(0,num_atms)))
        # Translate the atoms in mol1 to the atoms in mol2
        mol2_atom_sets = [ tuple([trans_dict[atm] for atm in bond_set]) for bond_set in mol1_atom_sets]
        # Add to list
        paired_atom_sets.append((mol1_atom_sets, mol2_atom_sets))
        # Check that the atom types are identical (test)
        mol1_atom_types = [ tuple([mol1.GetAtomWithIdx(atm).GetAtomicNum() for atm in bond_set]) for bond_set in mol1_atom_sets]
        mol2_atom_types = [ tuple([mol2.GetAtomWithIdx(atm).GetAtomicNum() for atm in bond_set]) for bond_set in mol2_atom_sets]
        assert mol1_atom_types == mol2_atom_types, "ATOM TYPES ARE NOT THE SAME ON THE DIHEDRAL ANGLE TO BE CALCULATED - THERE'S BEEN A MATCHING ERROR"
    # Return the list of lists of paired atoms between the structures
    return paired_atom_sets

def calculate_dihedral_angles(mol, dihedral_atom_sets):
    """find the dihedral angles of rotatable bonds in a molecule"""

    # Create list for the dihedrals (to be ordered in the same order as the input dihedral sets)
    dihedral_angles = []
    # Now calculate the dihedral angles between the sets identified previously
    conf = mol.GetConformer()
    # Loop through the angles => 2-3 is the rotatable bonds, 1,4 are the neighbours of 2,3 respectively
    for at1, at2, at3, at4 in dihedral_atom_sets:
        # Get the coordinates of the positions
        pos1 = conf.GetAtomPosition(at1)
        pos2 = conf.GetAtomPosition(at2)
        pos3 = conf.GetAtomPosition(at3)
        pos4 = conf.GetAtomPosition(at4)
        # Need to calculate three vectors 1->2, 2->3, 3->4
        vec1 = pos2 - pos1
        vec2 = pos3 - pos2
        vec3 = pos4 - pos3
        # Get the normals to the two planes (vec1-vec2 plane and vec2-vec3 plane))
        cross12 = vec1.CrossProduct(vec2)
        cross23 = vec2.CrossProduct(vec3)
        # Normalise the normals
        cross12.Normalize()
        cross23.Normalize()
        # Calculate dot-product and then inverse cosine to get the angle
        dot_prod = cross12.DotProduct(cross23)
        dihedral_rad = math.acos(dot_prod)
        dihedral_deg = 180*dihedral_rad/math.pi
        dihedral_angles.append(dihedral_deg)
    return dihedral_angles





from scitbx.array_family import flex
from scitbx.math import gamma_complete, gamma_incomplete
from scitbx.math import distributions

def test_significance_of_group_of_z_values(vals):
    """Test the significance of a group of z-values - assumes sum-squared is distributed as a chi-squared"""
    sum_sq = flex.sum_sq(flex.double(vals))
    n = len(vals)
    return chi_squared_test(sum_sq, n)

def chi_squared_test(sum_sq, n, max_iter=500):
    assert n < 344, 'Cannot test more than 343 values due to method of normalisig gamma function!'
    return 1.0-gamma_incomplete(a=n/2.0, x=sum_sq/2.0, max_iterations=max_iter)

def convert_pvalue_to_zscore(pval, two_tailed=True):
    """Convert a p-value to a z-score for a standard normal N(0,1)"""
    # If two-tailed test, need to halve the p-value
    if two_tailed:
        pval = pval/2.0
    # Create normal distribution to convert - N(0,1)
    nrm = distributions.normal_distribution()
    # Calculate the probability quantile (z-score) corresponding to 1-pval
    try:
        zsco = nrm.quantile(1.0-pval)
    except RuntimeError:
        # pval too small - return default (8.0 is max calculable)
        if pval < 0.00000000000000006:
            zsco = 8.2
        else:
            raise
    return zsco



import os, sys, copy, re

import libtbx.phil

import numpy
import pandas

from giant.xray.data import CrystalSummary
from giant.xray.data.cluster import CrystalGroup

#######################################

blank_arg_prepend = 'mtz='

master_phil = libtbx.phil.parse("""
input {
    mtz = None
        .type = path
        .multiple = True
    mtz_regex = None
        .type = str
}
check_for{
    label = None
        .type = str
        .multiple = True
}
output {
    out_dir = None
        .type = path
}
""")

#######################################

def run(params):

    if params.output.out_dir and (not os.path.exists(params.output.out_dir)):
        os.mkdir(params.output.out_dir)
    # Dump images into the current directory if no directory is given
    if not params.output.out_dir: img_dir = './'
    else: img_dir = params.output.out_dir

    #########################################################################################################

    if params.input.mtz_regex: labels = [re.findall(params.input.mtz_regex, m)[0] for m in params.input.mtz]
    else:                      labels = ['MTZ-{:06d}'.format(i) for i in range(len(params.input.mtz))]

    #########################################################################################################

    print '============================================>'
    print 'GROUPING BY SPACE GROUP'

    cg = CrystalGroup([CrystalSummary.from_mtz(mtz_file=f, id=lab) for f,lab in zip(params.input.mtz, labels)])

    #########################################################################################################

    print '==============================================================================>'
    print 'SPACE GROUPS: {} - {} DATASETS'.format(','.join(cg.space_groups), len(cg.crystals))

    #########################################################################################################

    res_sorted = sorted(cg.crystals, key=lambda c: c.high_res)
    high_res_crystal = res_sorted[0]
    low_res_crystal  = res_sorted[-1]

    print '===========================>'
    print 'HIGHEST RESOLUTION: {:.2f}'.format(high_res_crystal.high_res)
    print 'DATASET + PATH: {} - {}'.format(high_res_crystal.id, high_res_crystal.mtz_file)
    print 'LOWEST RESOLUTION: {:.2f}'.format(low_res_crystal.high_res)
    print 'DATASET + PATH: {} - {}'.format(low_res_crystal.id, low_res_crystal.mtz_file)

    vol_sorted = sorted(cg.crystals, key=lambda c: c.unit_cell.volume())
    high_vol_crystal = vol_sorted[-1]
    low_vol_crystal  = vol_sorted[0]

    print '===========================>'
    print 'HIGHEST VOLUME: {:.2f}'.format(high_vol_crystal.unit_cell.volume())
    print 'DATASET + PATH: {} - {}'.format(high_vol_crystal.id, high_vol_crystal.mtz_file)
    print 'LOWEST VOLUME: {:.2f}'.format(low_vol_crystal.unit_cell.volume())
    print 'DATASET + PATH: {} - {}'.format(low_vol_crystal.id, low_vol_crystal.mtz_file)

    #########################################################################################################

    for c in cg.crystals:

        if params.check_for.label:
            for lab in params.check_for.label:
                if lab not in c.column_labels:
                    print 'COLUMN NOT IN FILE ({}): {}'.format(lab, c.mtz_file)

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import os, sys, copy
import numpy
import iotbx.pdb
from scitbx.array_family import flex

def find_unused_chain_ids(hierarchies):
    unused_chain_ids = iotbx.pdb.systematic_chain_ids()
    # Extract current chain_ids
    current_chain_ids=[]; [current_chain_ids.extend([ch.id for ch in h.chains()]) for h in hierarchies]
    current_chain_ids=list(set(current_chain_ids))
    # Remove present chains
    for curr_id in current_chain_ids:   unused_chain_ids.remove(curr_id)
    return unused_chain_ids

def find_next_conformer_idx(hierarchy, all_ids=iotbx.pdb.systematic_chain_ids()):
    # Current altlocs in the ref structure
    current_conf_ids = sorted(list(hierarchy.altloc_indices()))
    # Find the next one that doesn't exist
    for i_id, new_conf_id in enumerate(all_ids):
        if new_conf_id not in current_conf_ids:
            return i_id

def resolve_residue_id_clashes(ref_hierarchy, mov_hierarchy, in_place=False):
    """Move residues in mov_hierarchiy to new chains if they have the same resid as a residue in ref_hierarchy but different resnames"""

    # Edit original copy or create new one?
    if (not in_place):
        ref_hierarchy = ref_hierarchy.deep_copy()
        mov_hierarchy = mov_hierarchy.deep_copy()

    # New chain to add clashing groups to
    new_chain = None
    new_chain_ids = find_unused_chain_ids(hierarchies=[ref_hierarchy, mov_hierarchy])

    # Find the residues with clashing resids
    residues_to_update = []
    for rg_mov in mov_hierarchy.residue_groups():
        # Extract label for this residue
        resid = rg_mov.resid(); chainid = rg_mov.parent().id
        rg_ref = [rg for rg in ref_hierarchy.residue_groups() if rg.resid()==resid and rg.parent().id==chainid]
        # MORE THAN ONE MATCHING RESIDUE -- ERROR?
        if len(rg_ref) > 1:     raise Exception('MORE THAN ONE MATCHING')
        # PRESENT ONLY IN MOVING
        elif len(rg_ref) == 0:  continue
        # PRESENT IN BOTH
        rg_ref = rg_ref[0]
        # Check to see if the residue is the same type as in the reference structure
        if map(str.strip, rg_ref.unique_resnames()) == map(str.strip,rg_mov.unique_resnames()):
            # Same residue -- that's fine
            continue
        else:
            print 'DIFFERENT RESIDUES WITH SAME ID - CHANGING CHAINS: {} != {}'.format(list(rg_ref.unique_resnames()), list(rg_mov.unique_resnames()))
            residues_to_update.append(rg_mov)

    # Go through and transfer the residue groups to new chains
    for rg_mov in residues_to_update:
        # Remove from old chain
        old_chain = rg_mov.parent()
        old_chain.remove_residue_group(rg_mov)
        # See if there is a residue with this id already present in the new chain
        if (not new_chain) or (rg_mov.resid() in new_chain.get_residue_ids()):
            new_chain = iotbx.pdb.hierarchy.chain(id=new_chain_ids.pop(0))
            # Add new chain to the same model as the old chain
            old_chain.parent().append_chain(new_chain)
        # Add to the new chain
        new_chain.append_residue_group(rg_mov)

    return mov_hierarchy

def compare_hierarchies(reference_hierarchy, query_hierarchy):
    """Find which residues are the same, different, or unique to each of the hierarchies"""

    # Can't be bothered to type those names all the time
    ref_h = reference_hierarchy; qry_h = query_hierarchy
    # Iterate through the reference hierarchy and find which residues have changed/are unique/are the same
    ref_only = []; ref_diff = []; ref_same = []

    # Iterate through residues
    for rg_ref in ref_h.residue_groups():
        # Check that the residue groups only have one residue name in each
        assert len(rg_ref.unique_resnames()) == 1, "Can't handle residues with more than one residue name"
        # Extract label for this residue
        resid = rg_ref.resid(); chainid = rg_ref.parent().id
        # Extract same rg for qry
        rg_qry = [rg for rg in qry_h.residue_groups() if rg.resid()==resid and rg.parent().id==chainid]

        # MORE THAN ONE MATCHING RESIDUE -- ERROR?
        if len(rg_qry) > 1:
            raise Exception('MORE THAN ONE MATCHING')
        # PRESENT ONLY IN REFERENCE
        elif len(rg_qry) == 0:
            ref_only.append((chainid, resid))
        # PRESENT IN BOTH
        elif len(rg_qry) == 1:
            # Check to see if the residue has moved
            rg_qry = rg_qry[0]
            # Check that the residue name is the same
            assert map(str.strip, rg_ref.unique_resnames()) == map(str.strip,rg_qry.unique_resnames()), 'Cannot Merge - Different Residues!: {} != {}'.format(list(rg_ref.unique_resnames()), list(rg_qry.unique_resnames()))
            # Extract atoms and check to see if the same
            rg_ref_ats = rg_ref.atoms()
            rg_qry_ats = rg_qry.atoms()
            # Check to see if the same length
            if len(rg_ref_ats) != len(rg_qry_ats):
                # There's been a change in the structure
                ref_diff.append((chainid, resid))
            else:
                # Extract coordinates for the atoms
                rg_ref_xyz = rg_ref_ats.extract_xyz()
                rg_qry_xyz = rg_qry_ats.extract_xyz()
                # Calculate the rmsd of the atoms (ASSUMING IN THE SAME ORDER)
                rmsd = (rg_ref_xyz - rg_qry_xyz).norm()
                # Check to see if the atoms have moved
                if rmsd != 0.0:
                    # There's been a change in the structure
                    ref_diff.append((chainid, resid))
                else:
                    # Same between the two structures
                    ref_same.append((chainid, resid))

    return ref_only, ref_diff, ref_same

def create_pure_alt_conf_from_proper_alt_conf(residue_group):
    """Take 'main conf' atoms and add to 'pure conformer' atom_groups"""

    main_ags = [ag for ag in residue_group.atom_groups() if ag.altloc=='']
    conf_ags = [ag for ag in residue_group.atom_groups() if ag.altloc!='']
    assert len(main_ags)==1, "Must be one atom_group in residue_group with altloc==''"
    assert len(conf_ags)!=0, "Must be at least one alternate conformer present"

    # Get a blank copy of the residue_group
    blank_rg = residue_group.detached_copy()
    for ag in blank_rg.atom_groups(): blank_rg.remove_atom_group(ag)
    assert not blank_rg.atom_groups(), blank_rg.atom_groups()

    # Atoms to be transferred to the other ags
    main_ag = main_ags[0]
    # Iterate through and add atoms from main_conf to each of the other atom_groups
    for conf_ag in conf_ags:
        new_conf_ag = conf_ag.detached_copy()
        new_main_ag = main_ag.detached_copy()
        # Set the occupancy of the main_conf atoms to be transferred
        max_occ = max(new_conf_ag.atoms().extract_occ())
        new_main_ag.atoms().set_occ(flex.double([max_occ]*new_main_ag.atoms().size()))
        # Copy atoms from the main conf to the alt conf
        for i_at, at in enumerate(new_main_ag.atoms()):
            new_conf_ag.insert_atom(i_at, at.detached_copy())
        # Add pure conf ag to the blank rg
        blank_rg.append_atom_group(new_conf_ag)
    return blank_rg

def merge_hierarchies(ref_hierarchy, mov_hierarchy):
    """Merge the hierarchies - creating new altlocs where necessary"""

    # Create new hierarchies to prevent editing the originals
    new_ref = ref_hierarchy.deep_copy()
    new_mov = mov_hierarchy.deep_copy()

    ######################################################################
    # FINDING NEW CONFORMER IDS
    ######################################################################

    # Find the next unused atlloc index to be used for the
    all_conformer_ids = iotbx.pdb.systematic_chain_ids()
    new_conformer_for_ref_idx = find_next_conformer_idx(hierarchy=ref_hierarchy, all_ids=all_conformer_ids)
    new_conformer_for_ref = all_conformer_ids[new_conformer_for_ref_idx]
    new_conformer_for_mov_idx = new_conformer_for_ref_idx + 1
    new_conformer_for_mov = all_conformer_ids[new_conformer_for_mov_idx]

    ######################################################################
    # IDENTIFYING WHICH RESIDUES ARE DIFFERENT BETWEEN MOVING AND REFERENCE STRUCTURES
    ######################################################################

    ref_only, ref_diff, ref_same = compare_hierarchies( reference_hierarchy = new_ref,
                                                        query_hierarchy     = new_mov   )
    print '=====================>>>'
    print '{!s} RESIDUES CONSERVED (SAME IN REFERENCE AND MOVING)'.format(len(ref_same))
    print '{!s} RESIDUES MOVED (DIFFERENT IN REFERENCE AND MOVING)'.format(len(ref_diff))
    print '{!s} RESIDUES UNIQUE TO REFERENCE CONFORMATION'.format(len(ref_only))

    ######################################################################
    # UPDATING CONFORMER LABELS FOR REFERENCE CONFORMERS
    ######################################################################

    # Record the number of conformers introduced into the ref structure
    conf_introduced = 0; conf_distributed = 0
    # Iterate through and update ref conformers
    for rg_ref in new_ref.residue_groups():
        # Extract label for this residue
        resid = rg_ref.resid(); chainid = rg_ref.parent().id
        # Major and mov are the same - Don't need to change the altloc
        if (chainid, resid) in ref_same:    continue
        # Major only or difference - Change the altloc for this residue to the defaults for the ref (if no conformers present)
        elif not rg_ref.have_conformers():
            assert len(rg_ref.atom_groups()) == 1
            rg_ref.atom_groups()[0].altloc = new_conformer_for_ref
            conf_introduced += 1
        # Conformers already present - make sure they are pure not proper
        elif '' in [ag.altloc for ag in rg_ref.atom_groups()]:
            new_rg = create_pure_alt_conf_from_proper_alt_conf(residue_group=rg_ref)
            # Find the rg's place in the parent chain
            ch = rg_ref.parent()
            rg_idx = ch.find_residue_group_index(rg_ref)
            # Remove the old rg and insert the new rg in it's place
            ch.remove_residue_group(rg_idx)
            ch.insert_residue_group(rg_idx, new_rg)
            conf_distributed += 1

    print '=====================>>>'
    print '{!s} CONFORMER IDS CREATED IN REFERENCE STRUCTURE'.format(conf_introduced)
    print '{!s} CONFORMER IDS UPDATED IN REFERENCE STRUCTURE'.format(conf_distributed)

    ######################################################################
    # UPDATING CONFORMER LABELS FOR MOVING CONFORMERS
    ######################################################################

    # If no conformer, change to default
    # If conformers, change to default + e.g. 'A'

    # Record the number of conformers introduced/incremented into the mov structure
    conf_introduced = 0; conf_incremented = 0
    # Iterate through and update mov conformers
    for rg_mov in new_mov.residue_groups():
        # Extract label for this residue
        resid = rg_mov.resid(); chainid = rg_mov.parent().id
        # If residue has not changed - skip as will not be transferred anyway
        if (chainid, resid) in ref_same:    continue
        # Transfer but no conformers - set to the default
        elif not rg_mov.have_conformers():
            assert len(rg_mov.atom_groups()) == 1
            rg_mov.atom_groups()[0].altloc = new_conformer_for_mov
            conf_introduced += 1
        # Conformers already present
        else:
            # Increment conformer letter
            for ag_min in rg_mov.atom_groups():
                if ag_min.altloc:
                    new_conf_idx = new_conformer_for_mov_idx + all_conformer_ids.index(ag_min.altloc) + 1
                    ag_min.altloc = all_conformer_ids[new_conf_idx]
            # Make sure they are pure not proper
            if '' in [ag.altloc for ag in rg_mov.atom_groups()]:
                new_rg = create_pure_alt_conf_from_proper_alt_conf(residue_group=rg_mov)
                # Find the rg's place in the parent chain
                ch = rg_mov.parent()
                rg_idx = ch.find_residue_group_index(rg_mov)
                # Remove the old rg and insert the new rg
                ch.remove_residue_group(rg_idx)
                ch.insert_residue_group(rg_idx, new_rg)
            conf_incremented += 1

    print '=====================>>>'
    print '{!s} CONFORMER IDS CREATED IN MOVING STRUCTURE'.format(conf_introduced)
    print '{!s} CONFORMER IDS UPDATED IN MOVING STRUCTURE'.format(conf_incremented)

    ######################################################################
    # TRANSFERRING RESIDUES FROM MOVING CONFORMATION
    ######################################################################

    # Create another copy of the ref conformation
    final_struct = new_ref.deep_copy()

    ref_chain_ids = [c.id for c in new_ref.chains()]

    # Iterate through chains
    for ch_min in new_mov.chains():

        # If not a matching chain, add to a new chain
        if ch_min.id not in ref_chain_ids:
            final_struct.models()[0].append_chain(ch_min.detached_copy())
        # Iterate through residue groups
        else:
            for rg_mov in ch_min.residue_groups():
                # Extract label for this residue
                resid = rg_mov.resid()
                chainid = rg_mov.parent().id
                # If residue has not changed - do not transfer
                if (chainid, resid) in ref_same:    continue
                # Find the matching residue group in the ref conformation
                rg_fin = [rg for rg in final_struct.residue_groups() if (resid==rg.resid() and chainid==rg.parent().id)]
                # MORE THAN ONE MATCHING RESIDUE -- ERROR?
                if len(rg_fin) > 1:     raise Exception('MORE THAN ONE MATCHING')
                # If not a matching residue group, add to a new residue group
                elif len(rg_fin)==0:
                    # Get the class name of the residue so that we can identify an appropriate chain object
                    res_class = [iotbx.pdb.common_residue_names_get_class(n) for n in rg_mov.unique_resnames()][0]
                    # Get the chains that already contain a residue with the same class
                    ch_fin = [c for c in final_struct.chains() if (res_class in c.get_residue_names_and_classes()[1].keys())]
                    # If can't find one, then just select all
                    if not ch_fin: ch_fin = final_struct.chains()
                    # Append the residue group to the first identified chain
                    ch_fin[0].append_residue_group(rg_mov.detached_copy())
                # One matching residue, so transfer the atom groups (don't have to worry about clashes as altlocs have already been changed)
                elif len(rg_fin)==1:
                    rg_fin = rg_fin[0]
                    # Transfer the atom groups
                    for min_ag in rg_mov.atom_groups(): rg_fin.append_atom_group(min_ag.detached_copy())

    return final_struct




import iotbx.pdb

def strip_pdb_to_string(input_pdb, remove_ter=False, remove_end=False):
    """Remove all ter cards from a pdb file. Output to string"""
    assert remove_ter or remove_end, 'No functions to perform'
    pdb_lines = open(input_pdb, 'r').readlines()
    out_lines = []
    for line in pdb_lines:
        if   remove_ter and (line[0:3]=='TER'):   pass
        elif remove_end and (line[0:3]=='END'):   pass
        else:   out_lines.append(line)
    return ''.join(out_lines)

def strip_pdb_to_input(input_pdb, remove_ter=False, remove_end=False):
    """Removes all of the TER and END records and regenerates them"""
    stripped = strip_pdb_to_string(input_pdb  = input_pdb,
                                        remove_ter = remove_ter,
                                        remove_end = remove_end     )
    return iotbx.pdb.hierarchy.input(pdb_string=stripped)

def strip_pdb_to_file(input_pdb, output_pdb=None, remove_ter=False, remove_end=False):
    """Remove all ter cards from a pdb file. If no output_file is given, operation is performed in_place"""
    if not output_pdb: output_pdb=input_pdb
    stipped = strip_pdb_to_string( input_pdb  = input_pdb,
                                        remove_ter = remove_ter,
                                        remove_end = remove_end     )
    with open(output_pdb, 'w') as fh: fh.write(stipped)

def reset_pdb_to_file(input_pdb, output_pdb=None, remove_ter=False, remove_end=False):
    """Removes all of the TER and END records and regenerates them and write to file"""
    if not output_pdb: output_pdb = input_pdb
    inp = strip_pdb_to_input(   input_pdb  = input_pdb,
                                remove_ter = remove_ter,
                                remove_end = remove_end   )
    inp.hierarchy.write_pdb_file(file_name=output_pdb)

def get_pdb_header(pdb_file):
    """Read the header of the pdb file (up to the atom records)"""

    contents = open(pdb_file, 'r').read().split('\n')
    marker = [i for i,l in enumerate(contents) if l.startswith('ATOM')][0]
    stripped = contents[:marker]
    return '\n'.join(stripped)+'\n'





import os, sys, glob, copy, time, itertools
import gtk
import pandas

from bamboo.plot import bar
from pandda.constants import PanddaInspectorFilenames, PanddaDatasetFilenames
from pandda.jiffies import pandda_summary
from pandda import inspect_html

def nonmodal_msg(msg):
    """Display an error window - non-modal"""
    d = gtk.MessageDialog(  type    = gtk.MESSAGE_INFO,
                            message_format = msg )
    d.show_all()
    while gtk.events_pending(): gtk.main_iteration(False)
    return d

def modal_msg(msg):
    """Display an error window - model"""
    d = gtk.MessageDialog(  type    = gtk.MESSAGE_INFO,
                            buttons = gtk.BUTTONS_CLOSE,
                            message_format = msg )
    d.run()
    d.destroy()

class PanddaEvent(object):
    def __init__(self, rank, info, top_dir):

        # Key for the pandas table
        self.index = info.name
        # Dataset tag
        self.dtag = info.name[0]
        # Dataset Information
        self.map_resolution  = round(info['analysed_resolution'],2)
        self.map_uncertainty = round(info['map_uncertainty'],2)
        self.rwork_rfree     = (round(info['r_work'],3),round(info['r_free'],3))
        # Event number for the dataset
        self.event_idx = int(info.name[1])
        # Position in the ranked list (1 -> n)
        self.rank = rank
        # Site Number (1 -> m)
        self.site_idx  = int(info['site_idx'])
        # Event Info
        self.est_1_bdc = round(info['1-BDC'], 2)
        # Z statistics
        self.z_peak = round(info['z_peak'], 1)
        self.z_mean = info['z_mean']
        self.cluster_size = int(info['cluster_size'])
        # Coordinate information
        self.ref_coords = info[['refx','refy','refz']]
        self.nat_coords = info[['x','y','z']]
        # Find the files for loading
        self.find_file_paths(top_dir=top_dir)

    def find_file_paths(self, top_dir):
        dtag = self.dtag
        # Top directory fo the pandda processing folder
        self.top_dir = top_dir
        # Identify other common directories
        self.stat_map_dir = os.path.join(top_dir, 'statistical_maps')
        self.ref_dir      = os.path.join(top_dir, 'reference')
        # Find the directory for the dataset
        self.dataset_dir  = glob.glob(os.path.join(top_dir, 'interesting_datasets/{!s}'.format(dtag)))
        assert len(self.dataset_dir) == 1, 'Non-unique dataset directory found: {!s}'.format(dtag)
        self.dataset_dir  = self.dataset_dir[0]

        # Identify dataset subdirectories and files
        self.ligand_dir = os.path.join(self.dataset_dir, 'ligand_files')
        self.model_dir  = os.path.join(self.dataset_dir, 'modelled_structures')

        if not os.path.exists(self.model_dir): os.mkdir(self.model_dir)

        # The most recent model of the protein in the pandda maps
        self.fitted_link        = os.path.join(self.model_dir,      PanddaDatasetFilenames.modelled_structure.format(dtag)  )
        # Unfitted model of the protein
        self.unfitted_model     = os.path.join(self.dataset_dir,    PanddaDatasetFilenames.aligned_structure.format(dtag)   )
        # Maps
        self.observed_map       = os.path.join(self.dataset_dir,    PanddaDatasetFilenames.sampled_map.format(dtag)         )
        self.z_map              = os.path.join(self.dataset_dir,    PanddaDatasetFilenames.z_map.format(dtag)               )
        self.mean_diff_map      = os.path.join(self.dataset_dir,    PanddaDatasetFilenames.mean_diff_map.format(dtag)       )
        self.event_map          = os.path.join(self.dataset_dir,    PanddaDatasetFilenames.event_map.format(dtag, self.event_idx, self.est_1_bdc))
        self.dataset_symmetry   = os.path.join(self.dataset_dir,    PanddaDatasetFilenames.symmetry_copies.format(dtag)     )

        # Ligand Files
        lig_files = glob.glob(os.path.join(self.ligand_dir, '*'))
        self.lig_pdbs = [f for f in lig_files if f.endswith('.pdb')]
        self.lig_cifs = [f for f in lig_files if f.endswith('.cif')]

    def find_current_fitted_model(self):
        """Get the most recent saved model of this protein"""
        fitted_outputs = sorted(glob.glob(os.path.join(self.model_dir, 'fitted-v*')))
        if fitted_outputs: return fitted_outputs[-1]
        else:              return None

    def find_new_fitted_model(self):
        """Get the path for the next model"""
        current = self.find_current_fitted_model()
        print 'CURRENT: {!s}'.format(current)
        if current: last_idx = int(current.replace('.pdb','')[-4:])
        else:       last_idx = 0
        new_fitted = 'fitted-v{:04d}.pdb'.format(last_idx+1)
        return os.path.join(self.model_dir, new_fitted)

    def write_fitted_model(self, protein_obj):
        new_fitted = self.find_new_fitted_model()
        # Save the fitted model to this file
        write_pdb_file(protein_obj, new_fitted)
        print '======================>'
        print 'SAVED PROTEIN MOL {!s} TO {!s}'.format(protein_obj, new_fitted)
        print '======================>'
        if os.path.exists(self.fitted_link):
#            assert os.path.islink(self.fitted_link), 'FILE IS NOT A LINK'
            # Delete the symbolic link
            os.unlink(self.fitted_link)
        print 'Linking {!s} -> {!s}'.format(new_fitted, self.fitted_link)
        # Create new link the most recent file
        os.symlink(os.path.basename(new_fitted), self.fitted_link)

#=========================================================================

class PanddaSiteTracker(object):
    events = None
    # Sites
    site_val = 1 # Site Value           1 -> m
    site_tot = 1 # Number of Sites      m
    # Ranks (Events)
    rank_idx = 0 # Indexing position    0 -> n-1
    rank_val = 1 # Rank Value           1 -> n
    rank_tot = 1 # Number of Events     n

    def __init__(self, parent, csv, top_dir):
        self.parent  = parent
        self.top_dir = top_dir
        self.parse_csv(csv)

    def parse_csv(self, csv):

        self.events = pandas.read_csv(csv, sep=',', dtype={'dtag':str})
        self.events = self.events.set_index(['dtag','event_idx'])
        print '====================>>>'
        print self.events
        print '====================>>>'
        current = self.get_new_current_event()
        print current
        print '====================>>>'
        self.update()

    def update(self):
        """Update position of the site list"""
        self.rank_val = self.rank_idx + 1
        self.rank_tot = len(self.events)
        self.site_val = int(self.events.iloc[self.rank_idx]['site_idx'])
        self.site_tot = len(set(self.events['site_idx']))

    def get_new_current_event(self):
        """Get an `Event` object for the current event"""
        self.update() # Ensure that we're up-to-date
        curr_event = self.events.iloc[self.rank_idx]
#        print '\n\nCurrent Event:\n\n{!s}\n\n'.format(curr_event)
        return PanddaEvent(rank=self.rank_val, info=curr_event, top_dir=self.top_dir)

    #-------------------------------------------------------------------------

    def at_first_event(self):
        self.update()
        return self.rank_val == 1

    def at_last_event(self):
        self.update()
        return self.rank_val == self.rank_tot

    #-------------------------------------------------------------------------

    def get_next(self):
        if self.rank_idx == self.rank_tot - 1:
            self.rank_idx = 0
#            return None
        else:
            self.rank_idx += 1
        return self.get_new_current_event()

    def get_prev(self):
        if self.rank_idx == 0:
            self.rank_idx = self.rank_tot - 1
#            return None
        else:
            self.rank_idx -= 1
        return self.get_new_current_event()

    def get_next_site(self):
        if self.site_val == self.site_tot:
            new_site_val = 1
#            return None
        else:
            new_site_val = self.site_val + 1
        self.rank_idx = self.events.index.get_loc(self.events[self.events['site_idx']==new_site_val].index[0])
        return self.get_new_current_event()

    def get_prev_site(self):
        if self.site_val == 1:
            new_site_val = self.site_tot
#            return None
        else:
            new_site_val = self.site_val - 1
        self.rank_idx = self.events.index.get_loc(self.events[self.events['site_idx']==new_site_val].index[0])
        return self.get_new_current_event()

#=========================================================================

class PanddaInspector(object):
    """Main Object in pandda.inspect"""

    def __init__(self, event_csv, site_csv, top_dir):

        self.log_table  = None
        self.site_table = None

        # List of events from pandda.analyse
        self.site_list = PanddaSiteTracker(parent=self, csv=event_csv, top_dir=top_dir)
        # Handling of coot commands
        self.coot = PanddaMolHandler(parent=self)

        # Working Directory
        self.top_dir = top_dir
        # Output csv from pandda.inspect
        self.output_event_csv = os.path.join(self.top_dir, 'analyses', PanddaInspectorFilenames.event_info)
        self.output_site_csv  = os.path.join(self.top_dir, 'analyses', PanddaInspectorFilenames.site_info)

        # Load previous data or create new table from site list, so we can record inspection data
        self.initialise_output_event_table( in_csv=event_csv, out_csv=self.output_event_csv )
        self.initialise_output_site_table(  in_csv=site_csv,  out_csv=self.output_site_csv  )
        # Save Tables
        self.write_output_csvs()

        # Object to hold the currently loaded event
        self.current_event = None

    def start_gui(self):
        self.gui = PanddaGUI(parent=self)
        self.gui.launch()

    def update_html(self):
        """Update various parts of the class - including output graphs"""

        # Post message in case this takes a long time...
        d = nonmodal_msg('Updating html output...')
        # Plot output graph of site list
        print 'Generating output graphs...'
        plot_vals = self.log_table['z_peak']
        view_vals = self.log_table['Viewed']
        modl_vals = self.log_table['Ligand Placed']
        colr_vals = ['limegreen' if m else 'red' if v else 'blue' for m,v in zip(modl_vals,view_vals)]
        site_idxs = self.log_table['site_idx']
        groups = [list(g[1]) for g in itertools.groupby(range(len(site_idxs)), key=lambda i: site_idxs[i])]
        bar.multiple_bar_plot_over_several_images(
                                f_template  = os.path.join(self.top_dir, 'results_summaries', PanddaInspectorFilenames.inspect_site_graph_mult),
                                plot_vals   = [[plot_vals[i] for i in g] for g in groups],
                                colour_vals = [[colr_vals[i] for i in g] for g in groups]   )
        # Write output html
        print 'Writing output html...'
        inspect_html.write_inspect_html(top_dir=self.top_dir, inspector=self)
        # Destroy update message
        d.destroy()

    def update_gui(self):

        # Update global position
        self.gui.labels['site_val'].set_label(str(self.site_list.site_val))
        self.gui.labels['site_tot'].set_label(str(self.site_list.site_tot))
        self.gui.labels['rank_val'].set_label(str(self.site_list.rank_val))
        self.gui.labels['rank_tot'].set_label(str(self.site_list.rank_tot))
        # Update current event and dataset information
        self.gui.labels['dtag'].set_label(str(self.current_event.dtag))
        self.gui.labels['e_idx'].set_label(str(self.current_event.event_idx))
        self.gui.labels['e_1_bdc'].set_label(str(self.current_event.est_1_bdc))
        self.gui.labels['zpeak'].set_label(str(round(self.current_event.z_peak,3)))
        self.gui.labels['csize'].set_label(str(self.current_event.cluster_size))
        self.gui.labels['map_res'].set_label(str(self.current_event.map_resolution))
        self.gui.labels['map_unc'].set_label(str(self.current_event.map_uncertainty))
        self.gui.labels['rwork_rfree'].set_label('{} / {}'.format(*self.current_event.rwork_rfree))

        # Reset the event comment boxes
        self.gui.objects['event comment text'].set_text(str(self.get_event_log_value('Comment')))

        # Reset the site comment boxes
        self.gui.objects['site name text'].set_text(str(self.site_table.get_value(index=self.current_event.site_idx, col='Name')))
        self.gui.objects['site comment text'].set_text(str(self.site_table.get_value(index=self.current_event.site_idx, col='Comment')))

        # Update the radio buttons - "Interesting"
        if self.get_event_log_value('Interesting') == True:
            self.gui.buttons['tp'].set_active(True)
        else:
            self.gui.buttons['fp'].set_active(True)
        # Update the radio buttons - "Ligand Placed"
        if self.get_event_log_value('Ligand Placed') == True:
            self.gui.buttons['placed'].set_active(True)
        else:
            self.gui.buttons['not placed'].set_active(True)
        # Update the radio buttons - "Ligand Confidence"
        if self.get_event_log_value('Ligand Confidence') == 'High':
            self.gui.buttons['high conf'].set_active(True)
        elif self.get_event_log_value('Ligand Confidence') == 'Medium':
            self.gui.buttons['med conf'].set_active(True)
        else:
            self.gui.buttons['low conf'].set_active(True)

        # Reset the merge button
        self.gui.buttons['merge'].child.set_text("Merge Ligand\nWith Model")

    #-------------------------------------------------------------------------

    def save_current(self):
        self.current_event.write_fitted_model(protein_obj=self.coot.open_mols['p'])
        self.write_output_csvs()

    def reset_current_to_last_model(self):
        close_molecule(self.coot.open_mols['p'])
        if os.path.exists(self.current_event.fitted_link):
            p = read_pdb(self.current_event.fitted_link)
        else:
            p = read_pdb(self.current_event.unfitted_model)
        self.coot.open_mols['p'] = p
        self.write_output_csvs()

    def reset_current_to_orig_model(self):
        close_molecule(self.coot.open_mols['p'])
        p = read_pdb(self.current_event.unfitted_model)
        self.coot.open_mols['p'] = p
        self.write_output_csvs()

    def load_new_event(self, new_event):
        if new_event is None:
            # TODO DO SOMETHING BETTER HERE TODO
            # Instead of loading new event, inform using the same model
            modal_msg(msg='Reloading Same MODEL (Unchanged)')
        else:
            self.current_event = self.coot.load_event(e=new_event)
        self.update_gui()

    def refresh_event(self):
        self.load_new_event(new_event=self.site_list.get_new_current_event())

    def load_next_event(self, skip_unmodelled=False, skip_viewed=False):
        if not self.check_valid_button_click(skip_unmodelled, skip_viewed): return
        new_event = self.site_list.get_next()
        # Loop until we get a modelled one (if requested)
        if (new_event is not None) and skip_unmodelled and (not os.path.exists(new_event.fitted_link)):
            if self.current_event.index == new_event.index:
                modal_msg(msg='No modelled datasets')
            else:
                print 'SKIPPING UNMODELLED: {} - {}'.format(new_event.dtag, new_event.event_idx)
                self.load_next_event(skip_unmodelled=skip_unmodelled, skip_viewed=skip_viewed)
            return
        # Loop until we get an unviewed one (if requested)
        if (new_event is not None) and skip_viewed and (self.log_table.get_value(index=new_event.index, col='Viewed') == True):
            print 'SKIPPING VIEWED: {} - {}'.format(new_event.dtag, new_event.event_idx)
            self.load_next_event(skip_unmodelled=skip_unmodelled, skip_viewed=skip_viewed)
            return
        # Actually load the event
        self.load_new_event(new_event=new_event)

    def load_prev_event(self, skip_unmodelled=False, skip_viewed=False):
        if not self.check_valid_button_click(skip_unmodelled, skip_viewed): return
        new_event = self.site_list.get_prev()
#        # Loop until we get a modelled one (if requested)
#        if (new_event is not None) and skip_unmodelled and (not os.path.exists(new_event.fitted_link)):
#            print 'SKIPPING UNMODELLED: {} - {}'.format(new_event.dtag, new_event.event_idx)
#            self.load_prev_event(skip_unmodelled=skip_unmodelled, skip_viewed=skip_viewed)
#            return
#        # Loop until we get an unviewed one (if requested)
#        if (new_event is not None) and skip_viewed and (self.log_table.get_value(index=new_event.index, col='Viewed') == True):
#            print 'SKIPPING VIEWED: {} - {}'.format(new_event.dtag, new_event.event_idx)
#            self.load_prev_event(skip_unmodelled=skip_unmodelled, skip_viewed=skip_viewed)
#            return
        # Actually load the event
        self.load_new_event(new_event=new_event)

    def load_next_site(self):
        self.load_new_event(new_event=self.site_list.get_next_site())

    def load_prev_site(self):
        self.load_new_event(new_event=self.site_list.get_prev_site())

    def load_dataset(self, dataset_id):
        """Find the next dataset with the given id"""
        new_event = self.site_list.get_next()
        # Check if this is the right dataset
        if (new_event is not None) and (new_event.dtag != dataset_id):
            if self.current_event.index == new_event.index:
                # Check if we've looped around
                modal_msg(msg='No dataset found for this id')
            else:
                # Load the next dataset if this doesn't match
                print 'Event does not match dataset id: {} - {}'.format(dataset_id, new_event.dtag)
                self.load_dataset(dataset_id=dataset_id)
            return
        # Actually load the event
        self.load_new_event(new_event=new_event)

    #-------------------------------------------------------------------------

    def check_valid_button_click(self, skip_unmodelled=None, skip_viewed=None):
        if skip_unmodelled == True:
            pass
        if skip_viewed == True:
            if sum(self.log_table['Viewed']) == len(self.log_table['Viewed']):
                modal_msg(msg='All models have been viewed')
                return False
        return True
    #-------------------------------------------------------------------------

    def initialise_output_event_table(self, in_csv, out_csv):
        """Read in the log table from pandda.analyse and merge with previous results form pandda.inspect"""

        # Read in the log table from pandda_analyse
        self.log_table = pandas.read_csv(in_csv, sep=',', dtype={'dtag':str})
        self.log_table = self.log_table.set_index(['dtag','event_idx'])
        # Create new columns (filled with blanks or defaults)
        self.log_table['Interesting'] = False
        self.log_table['Ligand Placed'] = False
        self.log_table['Ligand Confidence'] = 'Low'
        self.log_table['Comment'] = 'None'
        self.log_table['Viewed'] = False

        if os.path.exists(out_csv):
            print 'Merging with existing pandda_inspect_events.csv...'
            # Output csv already exists from previous run - reload and merge with in_csv
            inspect_prev = pandas.read_csv(out_csv, sep=',', dtype={'dtag':str})
            inspect_prev = inspect_prev.set_index(['dtag','event_idx'])
            # Merge with input table (only on the columns that should be updated)
            self.log_table.update(inspect_prev[['Interesting','Ligand Placed','Ligand Confidence','Comment','Viewed']])

    def initialise_output_site_table(self, in_csv, out_csv):
        """Read in the site definition table from pandda.analyse and merge with previous results form pandda.inspect"""

        # Read in the log table from pandda_analyse
        self.site_table = pandas.read_csv(in_csv, sep=',')
        self.site_table = self.site_table.set_index('site_idx')
        # Create new columns (filled with blanks for the moment)
        self.site_table['Name']     = 'None'
        self.site_table['Comment']  = 'None'

        if os.path.exists(out_csv):
            print 'Merging with existing pandda_inspect_events.csv...'
            # Output csv already exists from previous run - reload and merge with in_csv
            inspect_prev = pandas.read_csv(out_csv, sep=',')
            inspect_prev = inspect_prev.set_index('site_idx')
            # Merge with input table (only on the columns that should be updated)
            self.site_table.update(inspect_prev[['Name','Comment']])

    def get_event_log_value(self, col):
        return self.log_table.get_value(index=self.current_event.index, col=col)

    def set_event_log_value(self, col, value, write=True):
        self.log_table.set_value(index=self.current_event.index, col=col, value=value)
        print '====================>>>'
        print 'EVENT TABLE UPDATED'
        print '====================>>>'
        print 'Current Event:'
        print self.log_table.loc[self.current_event.index]
        print '====================>>>'
        if write: self.write_output_csvs()

    def get_site_log_value(self, col):
        return self.site_table.get_value(index=self.current_event.site_idx, col=col)

    def set_site_log_value(self, col, value, write=True):
        self.site_table.set_value(index=self.current_event.site_idx, col=col, value=value)
        print '====================>>>'
        print 'SITE TABLE UPDATED:'
        print '====================>>>'
        print 'Current Site:'
        print self.site_table.loc[self.current_event.site_idx]
        print '====================>>>'
        if write: self.write_output_csvs()

    def write_output_csvs(self):
        print 'Writing output csv: {}'.format(self.output_event_csv)
        self.log_table.to_csv(self.output_event_csv)
        print 'Writing output csv: {}'.format(self.output_site_csv)
        self.site_table.to_csv(self.output_site_csv)

#=========================================================================

class PanddaMolHandler(object):

    """Handles loaded Pandda Models (contains most of the coot functions)"""

    def __init__(self, parent):
        self.parent = parent
        self.open_mols = {}

    def close_all(self):
        for mol in molecule_number_list():
            close_molecule(mol)
        self.open_mols = {}

    def merge_ligand_with_protein(self):
        try: merge_molecules([self.open_mols['l']], self.open_mols['p'])
        except Exception as err: print err

    def move_ligand_here(self):
        try: move_molecule_to_screen_centre(self.open_mols['l'])
        except Exception as err: print err

    def load_event(self, e, close_all=True):
        """Load up all of the maps for an event"""

        # Close 'n' Load
        if close_all: self.close_all()

        # Re-centre camera
        set_rotation_centre(*e.ref_coords)

        # Load fitted version if it exists
        if os.path.exists(e.fitted_link): p = read_pdb(e.fitted_link)
        else:                             p = read_pdb(e.unfitted_model)

        # Load fitted version if it exists
        s = handle_read_ccp4_map(e.observed_map, 0)
        set_last_map_contour_level(1)
        set_map_displayed(s, 0)

        # Load fitted version if it exists
        z = handle_read_ccp4_map(e.z_map, 1)
        set_last_map_contour_level(3)
        set_map_displayed(z, 1)

        # Mean-Difference Map
        d = handle_read_ccp4_map(e.mean_diff_map, 1)
        set_last_map_contour_level_by_sigma(3)
        set_map_displayed(d, 0)

        # Event Map
        o = handle_read_ccp4_map(e.event_map, 0)
        set_last_map_contour_level(2*e.est_1_bdc)
        set_map_displayed(o, 1)

        # Symmetry contacts
        r = read_pdb(e.dataset_symmetry)
        set_mol_displayed(r, 0)

        # More Settings
        set_scrollable_map(o)
        set_imol_refinement_map(o)

        # Save mol numbers
        self.open_mols['p'] = p
        self.open_mols['s'] = s
        self.open_mols['z'] = z
        self.open_mols['d'] = d
        self.open_mols['o'] = o
        self.open_mols['r'] = r

        # Ligand Files
        if (len(e.lig_pdbs) == 1) and (len(e.lig_cifs) == 1):
            l_dict = read_cif_dictionary(e.lig_cifs[0])
            l = handle_read_draw_molecule_and_move_molecule_here(e.lig_pdbs[0])
            if os.path.exists(e.fitted_link): set_mol_displayed(l, 0)
            else:                             set_mol_displayed(l, 1)
            self.open_mols['l'] = l

            # Set the b-factors of the ligand to 20
            set_b_factor_molecule(l, 20)
            # Set the occupancy of the ligand to 2*(1-bdc)
            all_residue_ids = all_residues(l)
            if all_residue_ids:
                for res_chn, res_num, res_ins in all_residue_ids:
                    set_alt_conf_occ(l, res_chn, res_num, res_ins, [['', 2.0*e.est_1_bdc]])

        return e

class PanddaGUI(object):

    """GUI Class for pandda.inspect"""

    def __init__(self, parent):

        self.parent = parent

        # GUI objects
        self.labels = {}
        self.buttons = {}
        self.objects = {}

    def quit(self):
        self.store()
        self.parent.update_html()
        gtk.main_quit()

    def on_destroy(self,  widget=None, *data):
        self.quit()

    def on_delete(self, widget=None, *data):
        self.quit()

    def launch(self):
        """Launch GUI window"""

        # Create main window
        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.connect("delete_event", self.on_delete)
        self.window.connect("destroy_event", self.on_destroy)
        self.window.set_border_width(10)
        self.window.set_default_size(600, 400)
        self.window.set_title("PANDDA inspect")

        # Main VBox object for the window
        main_vbox = gtk.VBox(spacing=5)
        # -----------------------------------------------------
        hbox = gtk.HBox(spacing=5)
        # Create buttones to allow user to quit
        quit_buttons = self._quit_buttons()
        frame = gtk.Frame(); frame.add(quit_buttons)
        hbox.pack_start(frame)
        # Create progress summary table at the top of the window
        self.progress_table = self._progress_table()
        frame = gtk.Frame(); frame.add(self.progress_table)
        hbox.pack_start(frame)
        # Create buttons to navigate between datasets
        nav_buttons = self._navi_buttons_2()
        hbox.pack_start(nav_buttons)
        # Add to main vbox
        main_vbox.pack_start(hbox)
        # -----------------------------------------------------
        # Create buttons to navigate between datasets
        nav_buttons = self._navi_buttons_1()
        frame = gtk.Frame(); frame.add(nav_buttons)
        main_vbox.pack_start(frame)
        # -----------------------------------------------------
        hbox = gtk.HBox(homogeneous=False, spacing=5)
        # Create event summary table at the top of the window
        self.event_info_table = self._event_info_table()
        frame = gtk.Frame(); frame.add(self.event_info_table)
        hbox.pack_start(frame)
        # Create buttons to control the ligand
        lig_buttons = self._ligand_buttons()
        frame = gtk.Frame(); frame.add(lig_buttons)
        hbox.pack_start(frame)
        # Add to main vbox
        main_vbox.pack_start(hbox)
        # -----------------------------------------------------
        # Create buttones to record meta about the event
        rec_e_buttons = self._record_event_buttons()
        frame = gtk.Frame(); frame.add(rec_e_buttons)
        main_vbox.pack_start(frame)
        # -----------------------------------------------------
        # Create buttones to record meta about the event
        rec_s_buttons = self._record_site_buttons()
        frame = gtk.Frame(); frame.add(rec_s_buttons)
        main_vbox.pack_start(frame)

        # Link the buttons to the Inspector
        self.link_buttons()

        # Finally, show the window
        self.window.add(main_vbox)
        self.window.show_all()

        return self

    def link_buttons(self):
        """Link the buttons in the GUI to functions in PanddaInspector"""

        # Navigation buttons
        self.buttons['next'].connect("clicked", lambda x: [self.store(), self.parent.save_current(), self.parent.load_next_event()])
        self.buttons['prev'].connect("clicked", lambda x: [self.store(), self.parent.load_prev_event()])
        self.buttons['skip'].connect("clicked", lambda x: [self.store(), self.parent.load_next_event()])

        self.buttons['next-unviewed'].connect("clicked", lambda x: [self.store(), self.parent.load_next_event(skip_viewed=True)])
        self.buttons['next-modelled'].connect("clicked", lambda x: [self.store(), self.parent.load_next_event(skip_unmodelled=True)])
        self.buttons['next-site'].connect("clicked", lambda x: [self.store(), self.parent.load_next_site()])
        self.buttons['prev-site'].connect("clicked", lambda x: [self.store(), self.parent.load_prev_site()])

        self.buttons['go-to'].connect("clicked", lambda x: [self.store(), self.parent.load_dataset(dataset_id=self.objects['go-to-text'].get_text().strip()), self.objects['go-to-text'].set_text('')])

        # Quit
        self.buttons['quit'].connect("clicked", lambda x: [self.quit()])
        self.buttons['summary'].connect("clicked", lambda x: [self.store(), self.parent.update_html(), os.system('ccp4-python -Qnew -m pandda.jiffies.pandda_summary &')])
#        self.buttons['summary'].connect("clicked", lambda x: [self.store(), self.parent.update_html(), os.system('pandda.show_summary &')])
#        self.buttons['summary'].connect("clicked", lambda x: [self.store(), self.parent.update_html(), pandda_summary.run()])

        self.buttons['updatehtml'].connect("clicked", lambda x: [self.store(), self.parent.update_html()])

        # Structure Buttons
        self.buttons['save'].connect("clicked",  lambda x: self.parent.save_current())
        self.buttons['reload'].connect("clicked", lambda x: self.parent.reset_current_to_last_model())
        self.buttons['reset'].connect("clicked",  lambda x: self.parent.reset_current_to_orig_model())

        # Ligand Buttons
        self.buttons['merge'].connect("clicked", lambda x: [self.buttons['merge'].child.set_text('Already Merged\n(Click to Repeat)'), self.parent.coot.merge_ligand_with_protein()])
        self.buttons['move'].connect("clicked",  lambda x: self.parent.coot.move_ligand_here())

        # Meta Recording buttons
        self.buttons['tp'].connect("clicked",         lambda x: self.parent.set_event_log_value(col='Interesting', value=True))
        self.buttons['fp'].connect("clicked",         lambda x: self.parent.set_event_log_value(col='Interesting', value=False))
        self.buttons['high conf'].connect("clicked",  lambda x: self.parent.set_event_log_value(col='Ligand Confidence', value='High'))
        self.buttons['med conf'].connect("clicked",   lambda x: self.parent.set_event_log_value(col='Ligand Confidence', value='Medium'))
        self.buttons['low conf'].connect("clicked",   lambda x: self.parent.set_event_log_value(col='Ligand Confidence', value='Low'))
        self.buttons['placed'].connect("clicked",     lambda x: self.parent.set_event_log_value(col='Ligand Placed', value=True))
        self.buttons['not placed'].connect("clicked", lambda x: self.parent.set_event_log_value(col='Ligand Placed', value=False))

    def store(self):
        """Record information from the gui to the pandas table in the main object"""
        # Event records
        self.parent.set_event_log_value(col='Comment', value=self.objects['event comment text'].get_text(), write=False)
        self.parent.set_event_log_value(col='Viewed',  value=True, write=False)
        # Site records
        self.parent.set_site_log_value(col='Name',    value=self.objects['site name text'].get_text(),    write=False)
        self.parent.set_site_log_value(col='Comment', value=self.objects['site comment text'].get_text(), write=False)
        # Write csvs only once
        self.parent.write_output_csvs()

    def _navi_buttons_1(self):
        box = gtk.HBox(homogeneous=False, spacing=2)
        box.set_border_width(3)
        # ---
        b = gtk.Button(label="<<< Prev <<<\n(Don't Save Model)")
        b.child.set_justify(gtk.JUSTIFY_CENTER)
        self.buttons['prev'] = b
        box.pack_start(b)
        # ---
        b = gtk.Button(label=">>> Next >>>\n(Don't Save Model)")
        b.child.set_justify(gtk.JUSTIFY_CENTER)
        self.buttons['skip'] = b
        box.pack_start(b)
        # ---
        b = gtk.VSeparator()
        box.pack_start(b, expand=False, padding=5)
        # ---
        b = gtk.Button(label=">>> Next >>>\n(Save Model)")
        b.child.set_justify(gtk.JUSTIFY_CENTER)
        self.buttons['next'] = b
        box.pack_start(b)
        # ---
        return box

    def _navi_buttons_2(self):
        main_box  = gtk.VBox(homogeneous=True, spacing=5)
        # ---
        box1 = gtk.HBox(homogeneous=False, spacing=2)
        box1.set_border_width(3)
        frame = gtk.Frame(); frame.add(box1)
        main_box.pack_start(frame)
        # ---
        box2 = gtk.HBox(homogeneous=True, spacing=2)
        box2.set_border_width(3)
        frame = gtk.Frame(); frame.add(box2)
        main_box.pack_start(frame)
        # ---
        box3 = gtk.HBox(homogeneous=True, spacing=2)
        box3.set_border_width(3)
        frame = gtk.Frame(); frame.add(box3)
        main_box.pack_start(frame)
        # ---
        l = gtk.Label('Go to Dataset:')
        box1.pack_start(l, expand=False, fill=False, padding=5)
        # ---
        e = gtk.Entry(max=200)
        self.objects['go-to-text'] = e
        box1.pack_start(e, expand=True, fill=True, padding=5)
        # ---
        b = gtk.Button(label="Go")
        self.buttons['go-to'] = b
        box1.pack_start(b, expand=False, fill=False, padding=5)
        # ---
        b = gtk.Button(label="<<< Go to Prev Site <<<")
        self.buttons['prev-site'] = b
        box2.add(b)
        # ---
        b = gtk.Button(label=">>> Go to Next Site >>>")
        self.buttons['next-site'] = b
        box2.add(b)
        # ---
        b = gtk.Button(label=">>> Go to Next Unviewed >>>")
        self.buttons['next-unviewed'] = b
        box3.add(b)
        # ---
        b = gtk.Button(label=">>> Go to Next Modelled >>>")
        self.buttons['next-modelled'] = b
        box3.add(b)
        # ---
        return main_box

    def _ligand_buttons(self):
        box1 = gtk.VBox(homogeneous=True, spacing=2)
        box1.set_border_width(3)
        # ---
        b = gtk.Button(label="Merge Ligand\nWith Model")
        b.child.set_line_wrap(True)
        b.child.props.width_chars = 15
        b.child.set_justify(gtk.JUSTIFY_CENTER)
        self.buttons['merge'] = b
        box1.add(b)
        # ---
        b = gtk.Button(label="Move New Ligand Here")
        b.child.set_line_wrap(True)
        b.child.props.width_chars = 10
        b.child.set_justify(gtk.JUSTIFY_CENTER)
        self.buttons['move'] = b
        box1.add(b)
        # ---
        box2 = gtk.VBox(homogeneous=True, spacing=2)
        box2.set_border_width(3)
        # ---
        b = gtk.Button(label="Save Model")
        b.child.set_line_wrap(True)
        b.child.props.width_chars = 10
        b.child.set_justify(gtk.JUSTIFY_CENTER)
        self.buttons['save'] = b
        box2.add(b)
        # ---
        b = gtk.Button(label="Reload Last Saved Model")
        b.child.set_line_wrap(True)
        b.child.props.width_chars = 15
        b.child.set_justify(gtk.JUSTIFY_CENTER)
        self.buttons['reload'] = b
        box2.add(b)
        # ---
        b = gtk.Button(label="Reset to Unfitted Model")
        b.child.set_line_wrap(True)
        b.child.props.width_chars = 15
        b.child.set_justify(gtk.JUSTIFY_CENTER)
        self.buttons['reset'] = b
        box2.add(b)
        # ---
        hbox_main = gtk.HBox(spacing=5)
        hbox_main.pack_start(box1)
        hbox_main.pack_start(gtk.VSeparator(), expand=False, padding=5)
        hbox_main.pack_start(box2)

        return hbox_main

    def _record_event_buttons(self):
        # ---------------------------------------------
        hbox_1 = gtk.HBox(homogeneous=False, spacing=5)
        # ---
        hbox_1.pack_start(gtk.HBox(), expand=False, fill=False, padding=10)
        # ---
        l = gtk.Label('Event Comment:')
        hbox_1.pack_start(l, expand=False, fill=False, padding=5)
        # ---
        e = gtk.Entry(max=200)
        self.objects['event comment text'] = e
        hbox_1.pack_start(e, expand=True, fill=True, padding=5)
        # ---
        hbox_1.pack_start(gtk.HBox(), expand=False, fill=False, padding=10)
        # ---------------------------------------------
        hbox_2 = gtk.HBox(homogeneous=True, spacing=5)
        hbox_2.set_border_width(3)
        vbox_1_1 = gtk.VBox(homogeneous=True, spacing=2)
        vbox_1_2 = gtk.VBox(homogeneous=True, spacing=2)
        vbox_1_3 = gtk.VBox(homogeneous=True, spacing=2)
        hbox_2.add(vbox_1_1); hbox_2.add(vbox_1_2); hbox_2.add(vbox_1_3)
        # ---
        b = gtk.RadioButton(label="Mark Event as Interesting")
        self.buttons['tp'] = b
        vbox_1_1.add(b)
        # ---
        b = gtk.RadioButton(label="Mark Event as Not Interesting", group=b)
        self.buttons['fp'] = b
        vbox_1_1.add(b)
        # ---
        b = gtk.RadioButton(label="Ligand Placed")
        self.buttons['placed'] = b
        vbox_1_2.add(b)
        # ---
        b = gtk.RadioButton(label="No Ligand Placed", group=b)
        self.buttons['not placed'] = b
        vbox_1_2.add(b)
        # ---
        b = gtk.RadioButton(label="Model: High Confidence")
        self.buttons['high conf'] = b
        vbox_1_3.add(b)
        # ---
        b = gtk.RadioButton(label="Model: Medium Confidence", group=b)
        self.buttons['med conf'] = b
        vbox_1_3.add(b)
        # ---
        b = gtk.RadioButton(label="Model: Low Confidence", group=b)
        self.buttons['low conf'] = b
        vbox_1_3.add(b)
        # ---------------------------------------------
        vbox_main = gtk.VBox(spacing=0)
        vbox_main.pack_start(gtk.Label('Record Event Information (this event only)'), expand=False, fill=False, padding=5)
        vbox_main.pack_start(hbox_1, padding=0)
        vbox_main.pack_start(hbox_2, padding=5)

        return vbox_main

    def _record_site_buttons(self):
        # ---------------------------------------------
        hbox_1 = gtk.HBox(homogeneous=False, spacing=5)
        # ---
        hbox_1.pack_start(gtk.HBox(), expand=False, fill=False, padding=10)
        # ---
        l = gtk.Label('Name:')
        l.set_width_chars(10)
        hbox_1.pack_start(l, expand=False, fill=False, padding=5)
        # ---
        e = gtk.Entry(max=200)
        self.objects['site name text'] = e
        hbox_1.pack_start(e, expand=True, fill=True, padding=5)
        # ---
        hbox_1.pack_start(gtk.HBox(), expand=False, fill=False, padding=10)
        # ---------------------------------------------
        hbox_2 = gtk.HBox(homogeneous=False, spacing=5)
        # ---
        hbox_2.pack_start(gtk.HBox(), expand=False, fill=False, padding=10)
        # ---
        l = gtk.Label('Comment:')
        l.set_width_chars(10)
        hbox_2.pack_start(l, expand=False, fill=False, padding=5)
        # ---
        e = gtk.Entry(max=200)
        self.objects['site comment text'] = e
        hbox_2.pack_start(e, expand=True, fill=True, padding=5)
        # ---
        hbox_2.pack_start(gtk.HBox(), expand=False, fill=False, padding=10)
        # ---------------------------------------------
        vbox_main = gtk.VBox(spacing=0)
        vbox_main.pack_start(gtk.Label('Record Site Information (for all events with this site)'), expand=False, fill=False, padding=5)
        vbox_main.pack_start(hbox_1, padding=5)
        vbox_main.pack_start(hbox_2, padding=5)

        return vbox_main

    def _quit_buttons(self):
        # ---------------------------------------------
        vbox_1 = gtk.VBox(spacing=5)
        vbox_1.set_border_width(3)
        # ---
        b = gtk.Button(label="Quit")
        self.buttons['quit'] = b
        vbox_1.pack_start(b)
        # ---
        vbox_1.pack_start(gtk.HSeparator(), expand=False, padding=2)
        # ---
        b = gtk.Button(label="Summary")
        self.buttons['summary'] = b
        vbox_1.pack_start(b)
        # ---
        vbox_1.pack_start(gtk.HSeparator(), expand=False, padding=2)
        # ---
        b = gtk.Button(label="Update HTML")
        self.buttons['updatehtml'] = b
        vbox_1.pack_start(b)
        # ---
        hbox_main = gtk.HBox(spacing=5)
        hbox_main.pack_start(vbox_1)

        return hbox_main

    def _event_info_table(self):

        # Main box
        vbox_main = gtk.VBox()
        vbox_main.set_border_width(3)
        #  Pack sub-boxes
        hbox_sub_1 = gtk.HBox()
        hbox_sub_2 = gtk.HBox()
        vbox_main.pack_start(hbox_sub_1)
        vbox_main.pack_start(hbox_sub_2)

        ##############
        # HBOX SUB 1 #
        ##############

        # Dataset Name
        gtk_label = gtk.Label('Dataset ID')
        gtk_value = gtk.Label('None')
        gtk_box = gtk.EventBox(); gtk_box.add(gtk_value)
        hbox = gtk.HBox(homogeneous=True); hbox.add(gtk_label); hbox.add(gtk_box)
        frame = gtk.Frame(); frame.add(hbox)
        # Add to first box
        hbox_sub_1.pack_start(frame)
        # Store label to allow editing
        self.labels['dtag'] = gtk_value

        ##############
        # HBOX SUB 2 #
        ##############

        vbox_1 = gtk.VBox()
        vbox_2 = gtk.VBox()
        hbox_sub_2.pack_start(vbox_1)
        hbox_sub_2.pack_start(vbox_2)

        ##########
        # VBOX 1 #
        ##########

        # Create title
        title = gtk.Label('Event Information:')
        title.set_justify(gtk.JUSTIFY_LEFT)
        frame = gtk.Frame(); frame.add(title)
        # Add to first column
        vbox_1.pack_start(frame)

        # Event Number for Dataset
        gtk_label = gtk.Label('Event #')
        gtk_value = gtk.Label('-')
        gtk_box = gtk.EventBox(); gtk_box.add(gtk_value)
        hbox = gtk.HBox(homogeneous=True); hbox.add(gtk_label); hbox.add(gtk_box); hbox.set_border_width(3)
        frame = gtk.Frame(); frame.add(hbox)
        # Add to first column
        vbox_1.pack_start(frame)
        # Store label to allow editing
        self.labels['e_idx'] = gtk_value

        # Estimated Event Background Correction
        gtk_label = gtk.Label('1 - BDC')
        gtk_value = gtk.Label('-')
        gtk_box = gtk.EventBox(); gtk_box.add(gtk_value)
        hbox = gtk.HBox(homogeneous=True); hbox.add(gtk_label); hbox.add(gtk_box); hbox.set_border_width(3)
        frame = gtk.Frame(); frame.add(hbox)
        # Add to first column
        vbox_1.pack_start(frame)
        # Store label to allow editing
        self.labels['e_1_bdc'] = gtk_value

        # Z-Peak for Dataset
        gtk_label = gtk.Label('Z-blob Peak')
        gtk_value = gtk.Label('-')
        gtk_box = gtk.EventBox(); gtk_box.add(gtk_value)
        hbox = gtk.HBox(homogeneous=True); hbox.add(gtk_label); hbox.add(gtk_box); hbox.set_border_width(3)
        frame = gtk.Frame(); frame.add(hbox)
        # Add to first column
        vbox_1.pack_start(frame)
        # Store label to allow editing
        self.labels['zpeak'] = gtk_value

        # Z-Peak for Dataset
        gtk_label = gtk.Label('Z-blob Size')
        gtk_value = gtk.Label('-')
        gtk_box = gtk.EventBox(); gtk_box.add(gtk_value)
        hbox = gtk.HBox(homogeneous=True); hbox.add(gtk_label); hbox.add(gtk_box); hbox.set_border_width(3)
        frame = gtk.Frame(); frame.add(hbox)
        # Add to first column
        vbox_1.pack_start(frame)
        # Store label to allow editing
        self.labels['csize'] = gtk_value

        ##########
        # VBOX 2 #
        ##########

        # Create title
        title = gtk.Label('Dataset Information:')
        title.set_justify(gtk.JUSTIFY_LEFT)
        frame = gtk.Frame(); frame.add(title)
        # Add to second column
        vbox_2.pack_start(frame)

        # Resolution for Dataset
        gtk_label = gtk.Label('Resolution')
        gtk_value = gtk.Label('-')
        gtk_box = gtk.EventBox(); gtk_box.add(gtk_value)
        hbox = gtk.HBox(homogeneous=True); hbox.add(gtk_label); hbox.add(gtk_box); hbox.set_border_width(3)
        frame = gtk.Frame(); frame.add(hbox)
        # Add to second column
        vbox_2.pack_start(frame)
        # Store label to allow editing
        self.labels['map_res'] = gtk_value

        # Map Uncertainty for Dataset
        gtk_label = gtk.Label('Map Uncertainty')
        gtk_value = gtk.Label('-')
        gtk_box = gtk.EventBox(); gtk_box.add(gtk_value)
        hbox = gtk.HBox(homogeneous=True); hbox.add(gtk_label); hbox.add(gtk_box); hbox.set_border_width(3)
        frame = gtk.Frame(); frame.add(hbox)
        # Add to second column
        vbox_2.pack_start(frame)
        # Store label to allow editing
        self.labels['map_unc'] = gtk_value

        # R-Free/R-Work for Dataset
        gtk_label = gtk.Label('R-Free / R-Work')
        gtk_value = gtk.Label('-')
        gtk_box = gtk.EventBox(); gtk_box.add(gtk_value)
        hbox = gtk.HBox(homogeneous=True); hbox.add(gtk_label); hbox.add(gtk_box); hbox.set_border_width(3)
        frame = gtk.Frame(); frame.add(hbox)
        # Add to second column
        vbox_2.pack_start(frame)
        # Store label to allow editing
        self.labels['rwork_rfree'] = gtk_value

        # Currently Blank
        gtk_label = gtk.Label('-')
        gtk_value = gtk.Label('-')
        gtk_box = gtk.EventBox(); gtk_box.add(gtk_value)
        hbox = gtk.HBox(homogeneous=True); hbox.add(gtk_label); hbox.add(gtk_box); hbox.set_border_width(3)
        frame = gtk.Frame(); frame.add(hbox)
        # Add to second column
        vbox_2.pack_start(frame)
        # Store label to allow editing
        self.labels['blank'] = gtk_value

        return vbox_main

    def _progress_table(self):

        # First Column
        vbox_main = gtk.VBox(spacing=5)
        vbox_main.set_border_width(3)

        # Create title
        title = gtk.Label('Overall Inspection Event/Site Progress:')
        title.set_justify(gtk.JUSTIFY_LEFT)
        frame = gtk.Frame(); frame.add(title)
        # Add to first column
        vbox_main.pack_start(frame)

        # Event Number Name
        gtk_label_1 = gtk.Label('Event')
        gtk_value_1 = gtk.Label(0)
        gtk_label_2 = gtk.Label('of')
        gtk_value_2 = gtk.Label(0)
        # Add values to boxes
        gtk_box_1 = gtk.EventBox(); gtk_box_1.add(gtk_value_1)
        gtk_box_2 = gtk.EventBox(); gtk_box_2.add(gtk_value_2)
        hbox = gtk.HBox(homogeneous=True); hbox.add(gtk_label_1); hbox.add(gtk_box_1); hbox.add(gtk_label_2); hbox.add(gtk_box_2)
        frame = gtk.Frame(); frame.add(hbox)
        # Add to second column
        vbox_main.pack_start(frame)
        # Store label to allow editing
        self.labels['rank_val'] = gtk_value_1
        self.labels['rank_tot'] = gtk_value_2

        # Event Number Name
        gtk_label_1 = gtk.Label('Site')
        gtk_value_1 = gtk.Label(0)
        gtk_label_2 = gtk.Label('of')
        gtk_value_2 = gtk.Label(0)
        # Add values to boxes
        gtk_box_1 = gtk.EventBox(); gtk_box_1.add(gtk_value_1)
        gtk_box_2 = gtk.EventBox(); gtk_box_2.add(gtk_value_2)
        hbox = gtk.HBox(homogeneous=True); hbox.add(gtk_label_1); hbox.add(gtk_box_1); hbox.add(gtk_label_2); hbox.add(gtk_box_2)
        frame = gtk.Frame(); frame.add(hbox)
        # Add to second column
        vbox_main.pack_start(frame)
        # Store label to allow editing
        self.labels['site_val'] = gtk_value_1
        self.labels['site_tot'] = gtk_value_2

        return vbox_main

if __name__=='__main__':

#    import os
#    from pandda.inspect_main import PanddaInspector
    from pandda.constants import PanddaAnalyserFilenames

    #############################################################################################
    #
    # CHANGE COOT SETTINGS
    #
    #############################################################################################

    try:
        set_nomenclature_errors_on_read("ignore")
        set_colour_map_rotation_for_map(0)
        set_colour_map_rotation_on_read_pdb(0)
        set_recentre_on_read_pdb(0)
    except:
        pass

    #############################################################################################
    #
    # FIND THE INPUT CSV FILES
    #
    #############################################################################################

    work_dir = os.getcwd()
    hit_list = os.path.join(work_dir, 'analyses', PanddaAnalyserFilenames.event_info)
    site_csv = os.path.join(work_dir, 'analyses', PanddaAnalyserFilenames.site_info)

    #############################################################################################
    #
    # INITIALISE THE MAIN INSPECTOR OBJECT
    #
    #############################################################################################

    inspector = PanddaInspector(event_csv=hit_list, site_csv=site_csv, top_dir=work_dir)
    inspector.start_gui()
    inspector.refresh_event()




from bamboo.common.command import CommandManager
from bamboo.utils.mtz import MtzSummary
from bamboo.wrappers.coot.scripts import *

def validate_coot_script(script):
    """Performs checks on a coot script"""

    contents = open(script, 'r').read().strip()

    assert contents.endswith('coot_real_exit(1)'), 'Coot script does not EXIT! {!s}'.format(script)

    return

def run_coot(script, graphical=False, noguano=True):
    """Runs coot with the provided script"""

    # Check that the script is valid (e.g. causes coot to exit at the end)
    validate_coot_script(script)

    # Stop coot droppings?
    coot_flags = ['--no-guano']*(noguano)
    # Run with/without graphics?
    if graphical:
        coot_flags.extend(['-script',script])
    else:
        coot_flags.extend(['--no-graphics','-s',script])

    # Initialise
    COOT = CommandManager('coot')
    # Load arguments
    COOT.add_command_line_arguments(*coot_flags)
    # Run!
    COOT.run()

    return COOT




import math

# Code taken from Dimple (Marcin)

def calculate_view_quaternion(p1, p2):

    if p1 is None or p2 is None:
        return (0., 0., 0., 1.)
    d = (p2[0]-p1[0], p2[1]-p1[1], p2[2]-p1[2])
    length = math.sqrt(d[0]*d[0] + d[1]*d[1] + d[2]*d[2])
    d = (d[0]/length, d[1]/length, d[2]/length)
    # ref vector: 0 0 -1 (?)
    # cross product (a2 b3 - a3 b2, ..., ...)
    prod = (d[1], -d[0], 0)
    quat = (prod[0], prod[1], prod[2], 1-d[2])
    qlen = math.sqrt(sum(a*a for a in quat))
    return (quat[0]/qlen, quat[1]/qlen, quat[2]/qlen, quat[3]/qlen)

def multiply_quaternions(q1, q2):
    x, y, z, w = q1
    ax, ay, az, aw = q2
    return (w*ax + x*aw + y*az - z*ay,
            w*ay + y*aw + z*ax - x*az,
            w*az + z*aw + x*ay - y*ax,
            w*aw - x*ax - y*ay - z*az)




import os
from pandda.html import PANDDA_HTML_ENV
from giant.graphics import calculate_view_quaternion

# Get the template to be filled in
template = PANDDA_HTML_ENV.get_template('ccp4mg-pic.py')

point = [13.32,36.95,26.02]
com = [16.16,37.38,18.6]
orientation = calculate_view_quaternion(p1=com, p2=point)

ccp4mg_script = template.render({
                                    'view'  :{    'camera_centre' : [-1.0*p for p in point],
                                                  'orientation'   : orientation
                                                },
                                    'mol'   :{    'file'  : 'aligned_structure',
                                                  'path'  : './aligned.pdb',
                                                  'name'  : 'aligned_structure'
                                                },
                                    'map'   :{    'file'    : 'sampled_map',
                                                  'path'    : './sampled.ccp4',
                                                  'name'    : 'sampled_map',
                                                  'contour' : 1
                                                },
                                    'diff_map' :{ 'file'    : 'diff_map',
                                                  'path'    : './z_map.ccp4',
                                                  'name'    : 'diff_map',
                                                  'pos-contour' : 3,
                                                  'neg-contour' : -3
                                                }
                                })

print '################################'
print ccp4mg_script
print '################################'

outfile = './ccpm4_script.py'
assert not os.path.exists(outfile)

open(outfile, 'w').write(ccp4mg_script)














import os, sys

from bamboo.rdkit_utils.mol import check_pdb_readable

from rdkit import Chem

def get_centroid_from_file(model):
    """Get the centroid of an isolated ligand"""
    return get_centroid_from_mol(Chem.MolFromPDBFile(model))

def get_centroid_from_mol(mol):
    """Get the centroid of a mol"""

    num_atm = mol.GetNumHeavyAtoms()
    # Get the conformer (actual coords)
    conf = mol.GetConformer()
    # Get the atom objects
    atoms = mol.GetAtoms()
    # Zero the sums
    sumx, sumy, sumz, count = (0, 0, 0, 0)

    for a in atoms:
        # Skip Hydrogens (should have been removed anyway, but...)
        if a.GetAtomicNum() == 1:
            continue
        # Get the 3D coords of the atom
        pos = conf.GetAtomPosition(a.GetIdx())
        # Number of atoms used
        count += 1
        # Add the coords to the sum
        sumx += pos.x
        sumy += pos.y
        sumz += pos.z

    # Check we've got as many as we were expecting
    assert count == num_atm, 'Number of atoms used to calculate centroid != number of heavy atoms'

    return (sumx/count,sumy/count,sumz/count)

def calculate_coordinate_differences(model1, model2):
    """Calculate the differences between the atom coordinates of two identical structures"""

    # Read in mols and check for validity
    mol1 = check_pdb_readable(model1)
    mol2 = check_pdb_readable(model2)
    if (not mol1) or (not mol2):
        return None

    # Check that the mols are identical-ish
    if mol1.GetNumHeavyAtoms() != mol2.GetNumHeavyAtoms():
        raise EqualityError('Molecules are not identical (Num Atoms) {!s} != {!s}.\n{!s}\n{!s}'.format(mol1.GetNumHeavyAtoms(),mol2.GetNumHeavyAtoms(),Chem.MolToSmiles(mol1),Chem.MolToSmiles(mol2)))
    if mol1.GetNumBonds() != mol2.GetNumBonds():
        raise EqualityError('Molecules are not identical (Num Bonds) {!s} != {!s}:\n{!s}\n{!s}'.format(mol1.GetNumBonds(),mol2.GetNumBonds(),Chem.MolToSmiles(mol1), Chem.MolToSmiles(mol2)))

    # Gets atoms in mol1 (e.g. 14,5,3...) that match mol2 (1,2,3...)
    matchpatterns = mol1.GetSubstructMatches(mol2, uniquify=False)
    # Check to see if the molecules actually DO contain common substructures
    if not matchpatterns: return None

    differences = []
    # Get the conformers to access the coords
    conf1 = mol1.GetConformer(0)
    conf2 = mol2.GetConformer(0)
    # May be more than one matching pattern. Calculate all of them.
    for matchlist in matchpatterns:
        # reset the vector of coord difference for each match pattern
        match_diffs = []
        # idx2 = 0,1,2... idx1 = 14,5,3...
        for idx2, idx1 in enumerate(matchlist):
            # Get the atom coords
            atm1 = conf1.GetAtomPosition(idx1)
            atm2 = conf2.GetAtomPosition(idx2)
            # Append tuple of differences
            match_diffs.append((atm1.x - atm2.x, atm1.y - atm2.y, atm1.z - atm2.z))
        differences.append(match_diffs)
    # Return the differences corresponding to all of the ways of matching the molecules
    return differences

def get_atomic_equivalences(mol1, mol2):
    """Returns the list of atoms in mol1 that match with the atoms in mol2 => [pattern1, pattern2] ... pattern1 = [(mol1atm, mol2atm), ...]"""

    # Get the match patterns between mol1, mol2
    match_patterns = mol1.GetSubstructMatches(mol2, uniquify=False)

    # List of paired atoms in mol1 -> mol2
    atom_pairings = []

    if mol1.GetNumHeavyAtoms() != mol2.GetNumHeavyAtoms():
        raise EqualityError('Heavy Atom Numbers are not equal!')

    for match_pattern in match_patterns:
        atom_pairings.append(zip(match_pattern,range(0,mol1.GetNumHeavyAtoms())))

    return atom_pairings




import numpy

def normalise_modular_range(value, min, max):
    """Normalises values (e.g. angles) into the range {min, max}"""
    return numpy.mod(value-min, max-min)+min

def find_optimal_angle_window(angles, jump=5):
    """Find the optimal search window for the angles using a window of `jump`"""
    # The optimal shift for the std
    min_shift = min_std = 999
    # Shift the angles by an amount
    for shift in range(0, 360, jump):
        # Reset to {0,360}
        new_angles = numpy.mod(angles-shift, 360)
        # Calculate the std of this shift
        shift_std = numpy.std(new_angles)
        if shift_std < min_std:
            min_std = shift_std
            min_shift = shift
    return min_shift

def mean_std_angles(angles, interval=(-180,180)):
    """Calculate the mean and std of a set of angles. Returns mean in the range {0,360}"""
    # Find the optimal shift to avoid circularity (will fail for large stds)
    opt_shift = find_optimal_angle_window(angles)
    # Calculate the mean and standard deviation with the optimal shift
    ang_mean = numpy.mean(numpy.mod(angles-opt_shift, 360))+opt_shift
    ang_mean = normalise_modular_range(ang_mean, min=interval[0], max=interval[1])
    ang_std  = numpy.std(numpy.mod(angles-opt_shift, 360))
    return ang_mean, ang_std




from scitbx.array_family import flex

########################################################################################

def non_h(hierarchy, cache=None, copy=True):
    if not cache: cache=hierarchy.atom_selection_cache()
    sel = cache.selection('(not element H)')
    return hierarchy.select(sel, copy_atoms=copy)

def protein(hierarchy, cache=None, copy=True):
    if not cache: cache=hierarchy.atom_selection_cache()
    sel = cache.selection('(not element H) and pepnames')
    return hierarchy.select(sel, copy_atoms=copy)

def calphas(hierarchy, cache=None, copy=True):
    if not cache: cache=hierarchy.atom_selection_cache()
    sel = cache.selection('(not element H) and pepnames and (name CA)')
    return hierarchy.select(sel, copy_atoms=copy)

def backbone(hierarchy, cache=None, copy=True):
    if not cache: cache=hierarchy.atom_selection_cache()
    sel = cache.selection('(not element H) and pepnames and (name C or name CA or name N or name O)')
    return hierarchy.select(sel, copy_atoms=copy)

def sidechains(hierarchy, cache=None, copy=True):
    if not cache: cache=hierarchy.atom_selection_cache()
    sel = cache.selection('(not element H) and pepnames and not (name C or name CA or name N or name O)')
    return hierarchy.select(sel, copy_atoms=copy)

########################################################################################

def extract_backbone_atoms(residue):
    """Extract N, CA, C triplets for a residue"""
    residue_int = residue.residue_name_plus_atom_names_interpreter().atom_name_interpretation.expected
    residue_ats = residue.atoms().build_dict()
    n  = residue_ats[residue_int['N'][0]]
    ca = residue_ats[residue_int['CA'][0]]
    c  = residue_ats[residue_int['C'][0]]
    return (n, ca, c)

########################################################################################

#def get_backbone_atoms(atom_group):
#    """Get the backbone atoms for the atom_group"""
#
#    # Get the right interpreter for this residue
#    intr = iotbx.pdb.protein_atom_name_interpreters[atom_group.resname]
#    # Match the atom names
#    matches = intr.match_atom_names(atom_group.atoms().extract_name())
#    # Select the backbone atoms
#    c = matches.expected['C']
#
#    return

########################################################################################

# Deprecated
def get_calpha_sites(input_hierarchy):
    """Gets the coordinates of alpha carbons from the structure"""
    return flex.vec3_double([a.xyz for a in get_atom_selection(input_hierarchy, atom_names=[' CA '], atom=True, hetero=False)])

# Deprecated
def get_backbone_sites(input_hierarchy, cbeta=True):
    """Gets the coordinates of backbone atoms"""
    if cbeta: atoms = [' C  ',' N  ',' CA ',' O  ',' CB ']
    else:     atoms = [' C  ',' N  ',' CA ',' O  ']
    return flex.vec3_double([a.xyz for a in get_atom_selection(input_hierarchy, atom_names=[' C  ',' N  ',' CA ',' O  '], atom=True, hetero=False)])

########################################################################################

# Deprecated
def get_atom_selection(input_hierarchy, chain_names=None, res_names=None, atom_names=None, atom=True, hetero=False, proteinonly=True):
    """Iterate through atoms in input_hierarchy and select atoms matching the criteria"""

    if chain_names is None: chain_names = []
    if res_names   is None: res_names   = []
    if atom_names  is None: atom_names  = []

    filt_atoms = input_hierarchy.atoms_with_labels()

    # Select Hetero
    if (atom == True) and (hetero == True):
        filt_atoms = filt_atoms
    elif (atom == True) and (hetero == False):
        filt_atoms = [a for a in filt_atoms if a.hetero == False]
    elif (atom == False) and (hetero == True):
        filt_atoms = [a for a in filt_atoms if a.hetero == True]
    else:
        raise Exception('No Atoms Selected')

    # Get protein chains
    if proteinonly:
        prot_chains = [ch.id for ch in input_hierarchy.chains() if ch.is_protein()]
        filt_atoms = [a for a in filt_atoms if a.chain_id in prot_chains]

    # Get selected chains
    if chain_names:
        filt_atoms = [a for a in filt_atoms if a.chain_id in chain_names]

    # Get selected residues
    if res_names:
        filt_atoms = [a for a in filt_atoms if a.resname in res_names]

    # Get particular atoms
    if atom_names:
        filt_atoms = [a for a in filt_atoms if a.name in atom_names]

    return filt_atoms




import os, sys
import threading

from PyQt4.QtGui import QApplication
from pandda.summary import panddaWindow, panddaHtmlWidget

#######################################

blank_arg_prepend = None

master_phil = None

#######################################

def show_summary():
    app = QApplication(sys.argv)
    window = panddaWindow()
    window.add_tab(panddaHtmlWidget(name='Dataset Summary', content='./results_summaries/pandda_initial.html'))
    window.add_tab(panddaHtmlWidget(name='Results Summary', content='./results_summaries/pandda_analyse.html'))
    window.add_tab(panddaHtmlWidget(name='Inspect Summary', content='./results_summaries/pandda_inspect.html'))
    # Choose which window is initally open
    if   os.path.exists('./results_summaries/pandda_inspect.html'):
        window.setCurrentIndex(2)
    elif os.path.exists('./results_summaries/pandda_analyse.html'):
        window.setCurrentIndex(1)
    window.show()
    app.exec_()

#######################################

def run():
    t = threading.Thread(target=show_summary, args=())
    t.daemon = True
    t.start()
    t.join()

#######################################

if __name__=='__main__':
    run()




#:::::::::::::::::::::::::::::::::#
# ############################### #
# ###     Error Functions     ### #
# ############################### #
#:::::::::::::::::::::::::::::::::#

def throw_error(message):
    '''Print error message AND EXIT'''
    print "\n==========>"
    print "Terminal error"
    print "==========>"
    print "{!s}".format(message)
    print "==========>"
    print "QUITTING"
    print "==========>\n"
    raise SystemExit('Error Thrown')

def flag_error(message):
    '''Print error message AND DO NOT EXIT'''

    print "\n==========>"
    print "Non-terminal error"
    print "==========>"
    print "{!s}".format(message)
    print "==========>"
    print "CONTINUING"
    print "==========>\n"

class ErrorObject(object):
    """Object to record and report errors"""
    def __init__(self, verbose=False):
        self.verbose = verbose
        self.errorflags = {}
        self.warnings = []

    def __str__(self):
        """Formats error messages for printing"""
        return self.format_warnings()

    def print_warnings(self):
        if self.warnings:
            print(self.format_warnings())
        else:
            print('\n=====>\n=====> No Warnings!\n=====>\n')

    def write_warnings(self, filename):
        with open(filename,'w') as logfile:
            logfile.write(self.format_warnings())
        print('\n=====>\n=====> {!s} Warning(s) written to {!s}\n=====>\n'.format(len(self.warnings),filename))

    def format_warnings(self):
        """Formats the warning messages for printing"""
        warnings = ['\n=====>\n=====> {!s} Warning(s)\n=====>\n'.format(len(self.warnings))]
        warnings.extend(['\n {!s:>3} : {!s}'.format(i+1,warn) for i,warn in enumerate(self.warnings)])
        return ''.join(warnings)

    def record_warning(self, message):
        """Records any NON-TERMINAL error message"""
        self.warnings.append(message)
        if self.verbose:
            flag_error(message)

    def record_flag_warning(self, flag, message):
        """Records any NON-TERMINAL error message (with a flag)"""
        self.record_warning(message)
        if flag in self.errorflags:
            self.errorflags[flag].append(message)
        else:
            self.errorflags[flag] = [message]

    def record_error(self, message):
        """Raises any TERMINAL error message"""
        self.warnings.append(message)
        self.print_warnings()
        throw_error(message)

#:::::::::::::::::::::::::::::::::#
# ############################### #
# ###      Error Classes      ### #
# ############################### #
#:::::::::::::::::::::::::::::::::#

# General Errors

class ExternalProgramError(Exception):
    pass

# Pipline Errors

class PipelineError(Exception):
    pass

class OptionError(Exception):
    pass

class MolecularSubstitutionError(Exception):
    pass

class BuildingError(Exception):
    pass

class FittingError(Exception):
    pass

class RefinementError(Exception):
    pass

# Structure Errors

class MacroMolError(Exception):
    pass

class FragmentationError(Exception):
    pass

class DistanceError(Exception):
    pass

class ScoringError(Exception):
    pass

class AnalysisError(Exception):
    pass

# Ligand/Smile Errors

class LigandCheckError(Exception):
    pass

class SmileCheckError(Exception):
    pass

# RDkit Errors

class RDkitError(Exception):
    pass

class RDkitReadError(Exception):
    pass

class RDkitFragmentationError(Exception):
    pass

class BondAssignmentError(Exception):
    pass

# TODO Change this to MolEqualityError
class EqualityError(Exception):
    pass

# MTZ Errors

class ReflectionException(Exception):
    pass

class LabelError(Exception):
    pass

# Selection Errors

class RefinerSelectionError(Exception):
    pass

class LigandBuilderSelectionError(Exception):
    pass

class LigandFitterSelectionError(Exception):
    pass



import os, sys

PANDEMIC_TOP = os.path.join(__path__[0], '..', '..')
PANDEMIC_VERSION = 0.0.0
PANDEMIC_TEXT = """
------------------------------------------------------------------>
-
-  Pandemic Version {!s}
-
-  __________                    .___             .__
-  \\______   \\_____    ____    __| _/____   _____ |__| ____
-   |     ___/\\__  \\  /    \\  / __ |/ __ \\ /     \\|  |/ ___\\
-   |    |     / __ \\|   |  \\/ /_/ \\  ___/|  Y Y  \\  \\  \\___
-   |____|    (____  /___|  /\\____ |\\___  >__|_|  /__|\\___  >
-                  \\/     \\/      \\/    \\/      \\/        \\/
-
-  PAN-Dataset Ensemble Modelling of Isomorphous Crystals
-
-  When using this software, please cite:
-
-       Unpublished,
-       N Pearce, et al.
-
-  and include "PANDEMIC" in the KEYWRD section of related PDB files
-
------------------------------------------------------------------>
""".format(PANDEMIC_VERSION)

def welcome(user=None):
    """Welcome message"""

    if user is None:
        try:    user = os.getlogin()
        except: user = "Unknown user"

    try:
        from greetings import get_greeting
        print get_greeting(username=user)
    except SystemExit: raise
    except: pass

    try:
        print 'Hi {!s}. Welcome to Pandemic.'.format(user.upper())
    except: pass



import cctbx.sgtbx
import cctbx.uctbx

from scitbx.array_family import flex

from bamboo.common.masks import mask_collection
from bamboo.common import Meta

from pandda.handlers import DatasetHandler

########################################################################################################
#
#   HOLDER CLASSES
#
########################################################################################################

class MapHolder(object):
    """Class to hold map values and meta data"""
    def __init__(self, num, tag, map, unit_cell, space_group, meta, parent=None, child=None):
        assert isinstance(num, int), 'Num must be int. Type given: {!s}'.format(type(num))
        self.num = num
        assert isinstance(tag, str), 'Tag must be str. Type given: {!s}'.format(type(tag))
        self.tag = tag
        assert isinstance(map, flex.double), 'Map data must be flex.double. Type given: {!s}'.format(type(map))
        self.map = map
        assert isinstance(unit_cell, cctbx.uctbx.ext.unit_cell), 'Unit cell must be of type unit_cell. Type given: {!s}'.format(type(unit_cell))
        self.unit_cell = unit_cell
        assert isinstance(space_group, cctbx.sgtbx.ext.space_group), 'Space group must be of type space_group. Type given: {!s}'.format(type(space_group))
        self.space_group = space_group
        assert isinstance(meta, Meta), 'Meta must be dict. Type given: {!s}'.format(type(meta))
        self.meta = meta
        if parent:
            assert isinstance(parent, DatasetHandler) or isinstance(parent, MapHolder), 'parent must be of type DatasetHandler or MapHolder. Type given: {!s}'.format(type(parent))
            self.parent = parent
        if child:
            self.child = child

########################################################################################################
#
#   HOLDER LIST CLASSES
#
########################################################################################################

class HolderList(object):
    """Class for grouping many holders together"""
    _holder_class = None

    def __init__(self, *args):
        self._holder_list = []
        self._masks = mask_collection()
        # Allow for subclassing
        self.__class_init__(args)

    def __class_init__(self, args):
        pass

    def __getitem__(self, idx):
        if   isinstance(idx, int): return self.get(num=idx)
        elif isinstance(idx, str): return self.get(tag=idx)
        else: raise Exception('CANNOT INDEX EXCEPT BY int OR str. TYPE GIVEN: {!s}'.format(type(idx)))

    def __call__(self):
        """Return all holders"""
        return self.all()

    def all(self):
        """Return all holders"""
        return self._holder_list
    def all_nums(self):
        """Return the list of holder ids"""
        return [h.num for h in self.all()]
    def all_tags(self):
        """Return the list of holder tags"""
        return [h.tag for h in self.all()]

    def all_masks(self):
        """Return the mask object"""
        return self._masks
    def mask(self, mask_name, invert=False):
        """Retrieve a masked list of datasets"""
        if not mask_name:
            return self.all()
        else:
            return self._masks.mask(mask_name=mask_name, input_list=self.all(), invert=invert)

    def size(self, mask_name=None, invert=False):
        """Return the number of holders in the list (with optional mask applied)"""
        if mask_name:   return len(self.mask(mask_name=mask_name, invert=invert))
        else:           return len(self.all())

    def add(self, new_holders):
        """Add new datasets"""

        for new_h in new_holders:
            # Check all added datasets are the right class
            if self._holder_class:
                assert isinstance(new_h, self._holder_class), 'OBJECTS MUST BE OF TYPE: {!s}\n(ADDED OF TYPE: {!s})'.format(self._holder_class, type(new_h))
            # Check all added dataset id tags are strs
            assert isinstance(new_h.tag, str), 'TAG MUST BE str. Type given: {!s}'.format(type(new_h.tag))
            assert new_h.tag not in self.all_tags(), 'HOLDER WITH TAG ALREADY EXISTS: {!s}'.format(new_h.tag)
            # Check all added dataset id nums are ints
            assert isinstance(new_h.num, int), 'NUM MUST BE int. Type given: {!s}'.format(type(new_h.num))
            assert new_h.num not in self.all_nums(), 'HOLDER WITH NUM ALREADY EXISTS: {!s}'.format(new_h.num)
            # No problems, add to list
            self._holder_list.append(new_h)

    def get(self, tag=None, num=None):
        """Get a dataset by tag or num"""

        assert [num, tag].count(None) == 1, 'Must give EITHER num OR tag'
        if num: matching = [m for m in self.all() if m.num == num]
        else:   matching = [m for m in self.all() if m.tag == tag]
        if len(matching) == 0: raise Exception('NO MATCHING HOLDER FOUND - NUM: {!s}, TAG: {!s}'.format(num, tag))
        if len(matching) != 1: raise Exception('MORE THAN ONE MATCHING HOLDER FOUND - NUM: {!s}, TAG: {!s}'.format(num, tag))
        return matching[0]

########################################################################################################
#
#   HOLDER LIST SUBCLASSES
#
########################################################################################################

class DatasetHandlerList(HolderList):
    """Class for grouping many dataset handlers together"""
    _holder_class = DatasetHandler

    def __class_init__(self, *args):
        pass

class MapHolderList(HolderList):
    """Class for grouping many MapHolder objects together"""
    _holder_class = MapHolder

    def __class_init__(self, *args):
        pass





from bamboo.rdkit_utils.smile import find_structure_matches
from bamboo.wrappers.pdbutils import isolate_residue_by_res_id

def isolate_compound_from_file_by_smile(pdbin, pdbouttemplate, reference_smile, allow_part=True):
    """Isolates residues matching `reference_smile` from pdbin. Returns a list of files of the isolated residues based on pdbout"""

    if not pdbouttemplate.endswith('.pdb'):
        pdbouttemplate += '.pdb'
    pdbouttemplate = pdbouttemplate.replace('.pdb','.{!s}.pdb')

    # Load the molecule
    try:
        macro = MacroMol(pdbin)
    except MacroMolError:
        raise

    residues_to_remove = []

    # Get residues that match the input smile
    for res in macro.get_residues():
        if res.smile:
            full, part = find_structure_matches(res.smile, reference_smile)
            if full or (allow_part and part):
                residues_to_remove.append(res)

    if not residues_to_remove:
        raise Exception('Could Not Find Compound in File! ({!s} in {!s})'.format(reference_smile, pdbin))

    pdbout = []
    for i, res in enumerate(residues_to_remove):

        out = pdbouttemplate.format(i)
        pdbout.append(out)
        isolate_residue_by_res_id(pdbin, out, chain=res.chain, resnum=res.resnum, model='*', inscode=res.inscode)

    return pdbout




import os, sys, copy, re

import libtbx.phil
from bamboo.common.command import CommandManager
from giant.xray.data import CrystalSummary

#######################################

blank_arg_prepend = 'mtz='

master_phil = libtbx.phil.parse("""
input {
    mtz = None
        .type = path
        .multiple = True
    reference_mtz = None
        .type = path
}
options {
    free_r_flag = FreeR_flag
        .type = str
}
""")

#######################################

def run(params):

    assert params.input.reference_mtz is not None

    for mtz_file in params.input.mtz:

        reind_output_mtz = os.path.join(os.path.dirname(mtz_file), 'reind_'+os.path.basename(mtz_file))
        cad_output_mtz   = os.path.join(os.path.dirname(mtz_file), 'cad_'+os.path.basename(mtz_file))
        free_output_mtz  = os.path.join(os.path.dirname(mtz_file), 'free_'+os.path.basename(mtz_file))

        ######################################
        # REINDEX
        ######################################
        cm = CommandManager('pointless')
        cm.add_command_line_arguments([ 'hklin', mtz_file,
                                        'hklref', params.input.reference_mtz,
                                        'hklout', reind_output_mtz   ])
        cm.add_standard_input([ 'tolerance 5' ])
        cm.print_settings()
        cm.run()
        print '============================>'
        print cm.output
        print '============================>'
        print cm.error
        print '============================>'

        ######################################
        # Transfer Free-R Flags
        ######################################
        cm = CommandManager('cad')
        cm.add_command_line_arguments([ 'hklin1', reind_output_mtz,
                                        'hklin2', params.input.reference_mtz,
                                        'hklout', cad_output_mtz   ])
        cm.add_standard_input([ 'labin file_number 1 E1=IMEAN E2=SIGIMEAN',
                                'labin file_number 2 E1={}'.format(params.options.free_r_flag),
                                'labout file_number 2 E1=FreeR_flag',
                                'END'     ])
        cm.print_settings()
        cm.run()
        print '============================>'
        print cm.output
        print '============================>'
        print cm.error
        print '============================>'

        ######################################
        # Fill in missing reflections (nans)
        ######################################
        cm = CommandManager('uniqueify')
        cm.add_command_line_arguments([ '-f', 'FreeR_flag',
                                        cad_output_mtz,
                                        free_output_mtz   ])
        cm.print_settings()
        cm.run()
        print '============================>'
        print cm.output
        print '============================>'
        print cm.error
        print '============================>'

#        ######################################
#        # Complete Free-R Flags
#        ######################################
#        cm = CommandManager('freerflag')
#        cm.add_command_line_arguments([ 'hklin', cad_output_mtz,
#                                        'hklout', free_output_mtz   ])
#        cm.add_standard_input([ 'COMPLETE FREE=FreeR_flag',
#                                'END'     ])
#        cm.print_settings()
#        cm.run()
#        print '============================>'
#        print cm.output
#        print '============================>'
#        print cm.error
#        print '============================>'

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import mmtbx.f_model as model_handler
import iotbx.map_tools as map_tools

from cctbx import maptbx
from libtbx.math_utils import ifloor, iceil
from iotbx import crystal_symmetry_from_any
from scitbx.array_family import flex

def get_fft_map_from_f_obs_and_structure(f_obs, xray_structure, map_type='Fobs', d_min=None):
    """Create map from structure factors from array, phases from structure"""

    fmodel = model_handler.manager(f_obs=f_obs, xray_structure = xray_structure)
    density_map = fmodel.electron_density_map(update_f_part1=False)
    return density_map.map_coefficients(map_type).fft_map(d_min=d_min, symmetry_flags=maptbx.use_space_group_symmetry)

def write_1d_array_as_p1_map(file_name, map_data, grid_size, grid_spacing):
    """Take map data and write out to ccp4 map file - assumes the grid starts from the origin and the grid is cartesian!"""

    # Calculate the dimensions of the p1 box needed to enclose the map
    if isinstance(grid_spacing, tuple) or isinstance(grid_spacing, list):
        # (d1,d2,d3) -> non-isotropic grid spacing
        assert len(grid_spacing)==3, 'Grid spacing must be either a number or a tuple of length 3'
        dims_of_p1_box = tuple(flex.double(grid_size)*flex.double(grid_spacing))
    else:
        # d -> isotropic grid spacing
        dims_of_p1_box = tuple(flex.double(grid_size)*grid_spacing)
#    print 'Size of P1 Box:', dims_of_p1_box
    p1_box = generate_p1_box_from_size(dims_of_p1_box)
    # Copy and reshape
    map_data_to_write = map_data.deep_copy()
    map_data_to_write.reshape(flex.grid(grid_size))
    # Write map
    map_tools.write_ccp4_map(sites_cart=None, unit_cell=p1_box.unit_cell(), map_data=map_data_to_write, n_real=grid_size, file_name=file_name, buffer=0)

def generate_p1_box_from_sites(sites_cart, buffer=10.0) :
    xyz_max = sites_cart.max()
    a = xyz_max[0] + buffer
    b = xyz_max[1] + buffer
    c = xyz_max[2] + buffer
    return generate_p1_box_from_size((a,b,c))

def generate_p1_box_from_size(max_size):
    a,b,c = max_size
    combined = "%.3f,%.3f,%.3f,90,90,90,P1" % (a, b, c)
    symm = crystal_symmetry_from_any.from_string(combined)
    return symm




import os, urllib, datetime

from bamboo.constants import LONGEST_PEPTIDE, LONGEST_OLIGO, \
                                    AA_MODIFICATIONS_DICT, AA_3_TO_1_CODES_DICT, NUCLEOTIDE_DICT, \
                                    WATER_NAMES, ION_NAMES, SOLVENT_NAMES, \
                                    PDB_HEADERS_TO_KEEP, PDB_FOOTERS_TO_KEEP
from bamboo.rdkit_utils.smile import get_smile_from_block, match_smile_to_list

# RDKit Error Suppression
from rdkit import RDLogger
lg = RDLogger.logger()

class MacroMolError(Exception):
    pass

class MacroMol:
    """Class to handle a MacroMolecule"""

    def __init__(self, pdbin, keep_lines=-1, header_only=-1):
        """Initialise pdbfile <pdbin> object"""

        self.pdbin = pdbin
        self.pdbid = ''

        if os.path.exists(pdbin) and ('/' in pdbin or '\\' in pdbin or '.pdb' in pdbin):
            self.intype = 'file'
            self.pdbsource = pdbin
            self._set_state(1,0)
        elif len(pdbin)==4:
            self.intype = 'web'
            self.pdbsource = 'http://www.rcsb.org/pdb/files/{!s}.pdb'.format(pdbin)
            self._set_state(1,0)
        elif 'ATOM' in pdbin or 'HETATM' in pdbin:
            self.intype = 'block'
            self.pdbsource = 'block'
            self._set_state(1,0)
        else:
            raise TypeError("Invalid PDBIN given. Must be a '.pdb' file or a valid PDB code: \n\n{!s}".format(pdbin))
        # Update state
        self._set_state(keep_lines,header_only)

        # Read & Parse File
        self._read_pdb()
        self._parse_pdb()
        self._wipe_lines()

    def __iter__(self):
        return iter(self.get_chains())

    def __getitem__(self, id):
        if isinstance(id,int):
            return self.get_chains()[id]
        if isinstance(id,str):
            return self.get_chain(id)

    def _set_state(self, keep_lines, header_only):
        if keep_lines>=0: self.keep_lines=keep_lines
        if header_only>=0: self.header_only=header_only

    def _read_pdb(self):
        """Get and Read PDB information"""
        if self.intype=='file':
            self.pdb_lines = open(self.pdbsource,'r').readlines()
        elif self.intype=='web':
            self.pdb_lines = urllib.urlopen(self.pdbsource).readlines()
        elif self.intype=='block':
            self.pdb_lines = self.pdbin.replace('\n','\nZ\rX\rY\r').strip('Z\rX\rY\r').split('Z\rX\rY\r')
            self.pdbin = ''

        if len(self.pdb_lines) > 100000:
            raise MacroMolError('LARGE PDB FILE: {!s}'.format(self.pdbsource))

    def _parse_pdb(self):
        """Process PDB Information"""
        self.chain_list = []
        self.residue_list = []
        self.author_lines = []
        self.source_lines = []
        self.title_lines = []
        self.deposition_date = datetime.datetime.today()
        self.organism = ''
        self.experiment = ''
        # Record Lines that can be easily echoed to output
        self.easy_header = []
        self.easy_footer = []
        # Initialise current residue values
        curr_chain_id = None
        curr_chain = None
        curr_resnum = None
        curr_res = None
        # Build the lists by going through lines
        for line_num,line in enumerate(self.pdb_lines):
            fields = line.split()
            # Record easy-header and easy-footer
            if [flag for flag in PDB_HEADERS_TO_KEEP if line.startswith(flag)]:
                self.easy_header.append(line)
            if [flag for flag in PDB_FOOTERS_TO_KEEP if line.startswith(flag)]:
                self.easy_footer.append(line)
            # Parse meta-data
            if line[:4]!='ATOM' and line[:6]!='HETATM':
                # Source
                if line[:6]=='SOURCE':
                    self.source_lines.append(line)
                    if 'ORGANISM_COMMON' in line:
                        self.organism = line.split(':')[1].strip()
                # Authors
                if line[:6]=='AUTHOR':
                    self.author_lines.append(line)
                # Deposition Date - on first line of PDB entry
                if line[:6]=='HEADER':
                    try:
                        self.deposition_date = datetime.datetime.strptime(line[50:59],'%d-%b-%y')
                    except ValueError:
                        self.deposition_date = None
                # Release date - on REVDAT record, the one flagged "0" on column 31
                if line[:6]=='REVDAT':
                    if len(line)>31 and line[31]=='0':
                        self.release_date = datetime.datetime.strptime(line[13:22],'%d-%b-%y')
                    else:
                        self.release_date = None
                # Title
                if line[:5]=='TITLE':
                    self.title_lines.append(line)
                # Experiment type
                if line[12:27]=='EXPERIMENT TYPE':
                    self.experiment = line[45:50].lower().replace('-','')
                # (No other metadata just yet)
                continue
            # Parse atoms
            else:
                # Bail if headerOnly
                if self.header_only:
                    break
                # Parse res stats
                resname = line[17:20].strip()
                chain_id = line[21].strip()
                try:
                    resnum = int(line[22:26])
                except ValueError:
                    resnum = 0
                inscode = line[26].strip()
                # New chain as needed
                if not chain_id:
                    chain_id = ' '
                if chain_id != curr_chain_id:
                    if isinstance(curr_chain,Chain) and isinstance(curr_res,Residue):
                        # New Chain! - add the last residue to the old chain before moving on...
                        curr_chain.add_res(curr_res)
                        curr_res = None
                        curr_resnum = None
                    # Create a new chain
                    curr_chain_id=chain_id
                    if not self.chain_list:
                        curr_chain = Chain(chain_id)
                        self.chain_list.append(curr_chain)
                    elif chain_id not in self.get_chain_ids():
                        curr_chain = Chain(chain_id)
                        self.chain_list.append(curr_chain)
                    else:
                        curr_chain = self.get_chain(chain_id)
                # New residue if needed (incl. if new chain)
                if (resnum != curr_resnum) or (inscode != curr_inscode):
                    if isinstance(curr_res,Residue):
                        # New Residue - add the old one to the chain
                        curr_chain.add_res(curr_res)
                    # Change resnum and create a new residue
                    curr_resnum = resnum
                    curr_inscode = inscode
                    curr_res = Residue(resname,chain_id,resnum,inscode)
                    self.residue_list.append(curr_res)
                # Add line to residue
                curr_res.add_line(line,line_num)
        # At the end of the file, add the LAST Residue to the chain
        if isinstance(curr_chain,Chain) and isinstance(curr_res,Residue):
            curr_chain.add_res(curr_res)
        # Finish off
        for chn in self.chain_list:
            chn.update_types()

    def _wipe_lines(self):
        """Deletes pdb text if required"""
        if not self.keep_lines:
            print('Wiping...')
            self.pdb_lines = []

    def _force_full_read(self):
        """Force full parsing"""
        # Don't proceed if already in this loop
        if '_forcing' in self.__dict__ and self._forcing: return
        # Get the file lines if necessary, but then restore the state
        self._forcing=1
        if not self.chain_list:
            state=(self.keep_lines,self.header_only)
            self._set_state(-1,0)
            self._read_pdb()
            self._parse_pdb()
            self._wipe_lines()
            self._set_state(*state)
        self._forcing=0

    def get_chains(self):
        if not self.chain_list:
            self._force_full_read()
        return self.chain_list

    def get_chain(self, chain_id):
        """Returns <Chain> Object for <chainID>"""
        subset = [c for c in self.get_chains() if c.chain_id==chain_id]
        if len(subset)==1:
            return subset[0]
        else:
            raise Exception('More than one chain found with chainid: {!s}'.format(chain_id))

    def get_chain_ids(self):
        return [c.chain_id for c in self.get_chains()]

    def get_residues(self):
        return [res for chn in self.get_chains() for res in chn.residues]

    def get_residue(self, resid):
        residues = [res for res in self.get_residues() if res.get_res_id()==resid]
        if len(residues)>1: raise MacroMolError('Too Many Residues Found!')
        if len(residues)<1: raise KeyError('No Residue Found!')
        return residues[0]

    def get_aminos(self):
        return [res for chn in self.get_chains() for res in chn.aminos]

    def get_waters(self):
        return [res for chn in self.get_chains() for res in chn.waters]

    def get_ions(self):
        return [res for chn in self.get_chains() for res in chn.ions]

    def get_solvent(self):
        return [res for chn in self.get_chains() for res in chn.solvent]

    def get_unknowns(self):
        return [res for chn in self.get_chains() for res in chn.unknowns]

    def get_atoms(self):
        return [atm for res in self.get_residues() for atm in res.atoms]

    def print_chains(self):
        """Prints Chain IDs"""
        for ch in self.get_chains():
            print('{!s}: {!s}'.format(ch.chain_id, ch.types))

    def get_type(self):
        """Returns types of molecule present"""
        types_list = list(set([c_typ for c in self.get_chains() for c_typ in ['AMINO','NUCLEO'] if c_typ in c.types]))
        return ''.join(types_list)

    def is_protein(self):
        for chn in self.get_chains():
            if chn.is_protein():
                return True
        return False

    def is_protein_nucleo_complex(self):
        if ('AMINO' in self.get_type()) and ('NUCLEO' in self.get_type()):
            return True
        else:
            return False

    def get_title(self):
        return ''.join([l[10:].strip() for l in self.title_lines])

    def get_authors(self):
        return ''.join([l[10:].strip() for l in self.author_lines])

class Chain:
    """Class to handle PDB chains"""

    def __init__(self, chain_id=''):
        self.chain_id = chain_id.strip()
        self.residues = []
        self.aminos = []
        self.waters = []
        self.ions = []
        self.solvent = []
        self.unknowns = []
        self.types = []

    def __iter__(self):
        return iter(self.residues)

    def __getitem__(self, id):
        if isinstance(id,int):
            return self.residues[id]

    def __str__(self):
        str = 'Chain {!s}:\n\tNRes = {:d} ({:d} to {:d}), type {!s}'.format(self.chain_id, len(self.residues), self.first_resnum(), self.last_resnum(), '-'.join([t for t in self.types]))
        str += '\n\tSequence: {!s}'.format(self.get_sequence())
        if len(self.residues)<5:
            str += ' ==> '+''.join([n for n in self.get_names()])
        str += '\n\tAminos: {!s}'.format(len(self.aminos))
        str += '\n\tWaters: {!s}'.format(len(self.waters))
        str += '\n\tIons: {!s}'.format(len(self.ions))
        return str

    def __len__(self):
        return len(self.residues)

    def print_info(self):
        print(str(self))

    def add_res(self, residue):
        # Check the residue is up to date
        residue.assign_type_and_code()
        self.residues.append(residue)
        # Now sort it
        if residue.code == 'i':
            self.ions.append(residue)
        elif residue.code == 'o':
            self.waters.append(residue)
        elif 'SOLVENT' in residue.types:
            self.solvent.append(residue)
        elif 'UNKNOWN' in residue.types:
            self.unknowns.append(residue)
        else:
            self.aminos.append(residue)
        [self.types.append(t) for t in residue.types if t not in self.types]

    def get_res(self, resnum):
        subset = [r for r in self.residues if r.resnum == resnum]
        if len(subset)==1:
            return subset[0]
        else:
            raise Exception('More than one residue found with residue number {!s}'.format(resnum))

    def first_residue(self):
        if self.residues:
            return self.residues[0]

    def last_residue(self):
        if self.residues:
            return self.residues[-1]

    def first_resnum(self):
        return min(self.get_numbers())

    def last_resnum(self):
        return max(self.get_numbers())

    def get_numbers(self):
        return [r.resnum for r in self.residues]

    def get_names(self):
        return [r.resname for r in self.residues]

    def get_sequence(self):
        sequence = '-'
        for res in self.aminos:
            if len(res.code)==1:
                sequence += res.code
            else:
                if (sequence[-1]!='-'):
                    sequence += '-'
                sequence += '('+res.code+')-'
        return sequence.strip('-')

    def get_long_sequence(self):
        return '-'.join([r.resname for r in self.aminos])

    def update_types(self):
        [self.types.append(t) for r in self.residues for t in r.types if t not in self.types]

    def is_protein(self):
        if self.get_aa_count() > LONGEST_PEPTIDE:
            return True
        else:
            return False

    def is_peptide(self):
        aa_count = self.get_aa_count()
        if aa_count > 0 and aa_count <= LONGEST_PEPTIDE:
            return True
        else:
            return False

    def get_aa_count(self):
        return len([r for r in self.residues if (('AMINO' in r.types) or ('AMINO-MOD' in r.types))])

class Residue:
    """Class to handle PDB residues"""

    def __init__(self, resname, chain, resnum, inscode):
        self.resname = resname.strip()
        self.chain = chain.strip()
        self.resnum = resnum
        self.inscode = inscode
        self.first_line = -1
        self.last_line = -1
        self.atoms = []
        self.smile = ''
        self.assign_type_and_code()

    def __iter__(self):
        return iter(self.atoms)

    def __len__(self):
        return len(self.atoms)

    def get_pdb_string(self):
        return ''.join([atm.line for atm in self.atoms])

    def print_lines(self):
        print(self.get_pdb_string())

    def add_line(self, pdb_line, line_num):
        pdb_line = pdb_line.strip('\n')
        # Make the right length for PDB format
        if len(pdb_line)<80:
            pdb_line = pdb_line + ' '*(80-len(pdb_line))
        # Add new atom
        self.atoms.append(Atom(pdb_line))
        # Check for line numbers
        if self.first_line<0:
            self.first_line=line_num
        self.last_line=line_num

    def assign_type_and_code(self):
        types = []
        alt_codes = []
        # Amino Acids
        if self.resname in AA_3_TO_1_CODES_DICT:
            types.append('AMINO')
            code = AA_3_TO_1_CODES_DICT[self.resname]
        # Nucleotides
        elif self.resname in NUCLEOTIDE_DICT:
            types.append('NUCLEO')
            code = NUCLEOTIDE_DICT[self.resname]
        # Waters
        elif self.is_water():
            types.append('WATER')
            code = 'o'
        # Ions
        elif self.is_ion():
            types.append('ION')
            code = 'i'
        # Solvent/Buffer
        elif self.is_solvent_or_buffer():
            types.append('SOLVENT')
            code = self.resname
            self.get_smiles()
        # Modified amino acids or ligands (or unknown)
        else:
            code = self.resname
            self.get_smiles()
            if self.smile:
                matches = match_smile_to_list(self.smile, AA_MODIFICATIONS_DICT.keys(), assignbonds=True)
                if matches:
                    types.append('AMINO-MOD')
                    [alt_codes.extend(AA_MODIFICATIONS_DICT[s]) for s in matches]
            if not (types and code):
                types.append('UNKNOWN')

        self.types = types
        self.code = code
        if alt_codes:
            self.alt_codes = alt_codes
        else:
            self.alt_codes = [code]

    def is_water(self):
        if self.resname.upper() in WATER_NAMES:
            return True
        else:
            return False

    def is_ion(self):
        """Test if the residue is a non-water ion"""

        # Count number of atoms (accounting for differing alternate locations)
        if len([at for at in self.atoms if at.altloc in ['', 'A']])!=1:
            return False
        elif self.is_water():
            return False
        else:
            return True

    def is_solvent_or_buffer(self):
        """Test is the residue is a (common) solvent or buffer molecule"""

        # Count number of atoms (must be greater than one)
        if len([at for at in self.atoms if at.altloc in ['', 'A']])==1:
            return False
        elif self.resname.upper() in SOLVENT_NAMES:
            return True
        else:
            return False

    def get_smiles(self):
        if self.atoms:
            lg.setLevel(RDLogger.CRITICAL)

            altlocs = sorted(set([at.altloc for at in self.atoms if at.altloc]))
            for alt in altlocs:
                atms = [atm for atm in self.atoms if ((not atm.altloc) or (atm.altloc == alt))]
                pdb_string = ''.join([atm.line[0:16]+' '+atm.line[17:] for atm in atms])
                smile = get_smile_from_block(pdb_string)
                if smile:
                    self.smile = smile
                    break
#            self.smile = get_smile_from_block(self.get_pdb_string())
            lg.setLevel(RDLogger.ERROR)

    def get_res_id(self):
        """Get a tuple that identifies the residue (used as a key in other functions to refer to the residue)"""
        return (self.resname, self.chain, self.resnum, self.inscode)

    def get_non_h_atoms(self):
        """Get non-hydrogen atoms of the residue"""
        if not self.atoms:
            raise ValueError('No atoms in residue! Cannot calculate average occupancy!')
        return [atm for atm in self.atoms if atm.element != 'H']

    def get_centroid(self):
        """Return the centre of the residue, giving equal weight to each of the atoms (no H)"""
        non_h = self.get_non_h_atoms()
        x_cent = sum([atm.x for atm in non_h])/len(non_h)
        y_cent = sum([atm.y for atm in non_h])/len(non_h)
        z_cent = sum([atm.z for atm in non_h])/len(non_h)
        return (x_cent, y_cent, z_cent)

    def get_mean_occupancy(self):
        """Get the average occupancy for atoms in the residue (excluding hydrogens)"""
        if not self.atoms:
            raise ValueError('No atoms in residue! Cannot calculate average occupancy!')
        non_h = self.get_non_h_atoms()
        mean_occupancy = sum([atm.occupancy for atm in non_h])/len(non_h)
        return mean_occupancy

    def get_mean_bfactor(self):
        """Get the average bfactor for atoms in the residue (excluding hydrogens)"""
        if not self.atoms:
            raise ValueError('No atoms in residue! Cannot calculate average bfactor!')
        non_h = self.get_non_h_atoms()
        mean_bfactor = sum([atm.bfactor for atm in non_h])/len(non_h)
        return mean_bfactor

    def update_inscode(self, inscode):
        """Change the inscode of atom and update atom line"""
        if not self.atoms:
            raise ValueError('No atoms in residue! Cannot change insertion code!')
        new_ins = str(inscode.upper())
        assert len(new_ins)<2, 'New inscode too long! {!s}'.format(new_ins)
        for atm in self.atoms:
            # Set the inscode
            atm.inscode = new_ins
            # Change the pdbline
            newline = list(atm.line)
            newline[26] = new_ins
            atm.line = ''.join(newline)
            # Sanity check line length
            assert len(atm.line)==81, 'ATOM LINE IS NOT THE RIGHT LENGTH FOR PDB FORMAT! {!s}'.format(atm.line)

    def update_atom_numbers(self,atom_numbering):
        """Update the atom numbers to those in atom_numbering - Requires a dict with atom names as the keys for the new numbers"""
        if not self.atoms:
            raise ValueError('No atoms in residue! Cannot change atom numbers!')
        for atm in self.atoms:
            # Set the new serial num
            atm.serialnum = atom_numbering[atm.atomname]
            # Change the pdbline
            newline = list(atm.line)
            newline[6:11] = '{:>5}'.format(str(atom_numbering[atm.atomname]))
            atm.line = ''.join(newline)
            # Sanity check line length
            assert len(atm.line)==81, 'ATOM LINE IS NOT THE RIGHT LENGTH FOR PDB FORMAT! {!s}'.format(atm.line)

class Atom:
    """Atom Class"""

    def __init__(self, line):

        # Process input line and populate record
        line = line.strip('\n').upper()
        # Make the right length
        if len(line) < 80:
            line = line+' '*(80-len(line))
        # Save the line
        self.line = line+'\n'
        # Check the ATOM or HETATM records
        try:
            self.recordtype = line[0:6].strip()
            self.serialnum  = line[6:11].strip()
            self.atomname   = line[12:16].strip()
            self.altloc     = line[16].strip()
            self.resname    = line[17:20].strip()
            self.chnname    = line[21].strip()
            self.resnum     = line[22:26].strip()
            self.inscode    = line[26].strip()
            self.x          = float(line[30:38])
            self.y          = float(line[38:46])
            self.z          = float(line[46:54])
            self.occupancy  = float(line[54:60])
            self.bfactor    = float(line[60:66])
            self.segment    = line[72:76].strip()
            self.element    = line[76:78].strip()
            self.charge     = line[78:80].strip()
        except ValueError as Err:
            raise MacroMolError('Error in PDB File Atom Coordinates. Maybe Wrong File Format?\nOriginal Error is: {!s}'.format(Err))




import os

import scipy, numpy

from scitbx.array_family import flex

import iotbx.pdb
from cctbx import crystal, sgtbx

from giant.stats.cluster import find_connected_groups

def match_sites_by_symmetry(ref_sites, query_sites, unit_cell, space_group, cutoff=5):
    """Pair sites in query_sites to sites in ref_sites, allowing for symmetry"""

    pairings = numpy.zeros((len(ref_sites), len(query_sites)), dtype=numpy.int)

    for i_ref, ref in enumerate(ref_sites):
        ref_frac = unit_cell.fractionalize(ref)
        sym_sites_ref = sgtbx.sym_equiv_sites(space_group   = space_group,
                                              unit_cell     = unit_cell,
                                              original_site = ref_frac  )

        for i_query, query in enumerate(query_sites):
            query_frac = unit_cell.fractionalize(query)
            min_dist = sgtbx.min_sym_equiv_distance_info(sym_sites_ref, query_frac).dist()
            if min_dist < cutoff:
                pairings[i_ref, i_query] = 1

    return pairings

def combine_hierarchies(list_of_hierarchies):
    """Combine a list of hierarchies into one hierarchy -- Requires all of the chain identifiers to be unique"""
    top_h = list_of_hierarchies[0].deep_copy()
    for next_h in list_of_hierarchies[1:]: top_h.transfer_chains_from_other(next_h.deep_copy())
    return top_h

def find_symmetry_equivalent_groups(points_frac, sym_ops, unit_cell, cutoff_cart):
    """Group sets of points by minimum distance between points, allowing for symmetry"""

    # Pre-calculate the square distance
    dist_cut_sq = cutoff_cart**2
    # Matrix for whether they are near to each other
    equiv_sites = numpy.zeros([len(sym_ops), len(points_frac), len(points_frac)], dtype=int)

    # Apply the symmetry operations to each group to see if it is near to other groups
    for i_sym_op, sym_op in enumerate(sym_ops):
        # Transformed groups under this symmetry operation
        trans_points_frac = [sym_op.as_rational().as_float() * pf for pf in points_frac]
        # Loop through original points
        for i_group_1, group_1 in enumerate(points_frac):
            # Loop through transformed points again
            for i_group_2, group_2 in enumerate(trans_points_frac):
                # Comparing group to itself - skip
                if i_group_1 == i_group_2:
                    equivalent = True
                else:
                    # Start by assuming the clusters are not related
                    equivalent = False
                    # Loop through and see if the clusters overlap
                    for qp in group_2:
                        # Calculate difference and convert back to cartesian
                        diffs_cart = unit_cell.orthogonalize(group_1 - qp)
                        # Check if any are closer than the minimum required
                        if min(diffs_cart.dot()) < dist_cut_sq:
                            equivalent = True
                            break
                # If the clusters overlap, they are equivalent
                if equivalent:
                    equiv_sites[(i_sym_op, i_group_1, i_group_2)] = 1

    # Condense the cluster equivalence - take max over the symmetry operations and group by connected paths
    sym_groups = find_connected_groups(connection_matrix=equiv_sites.max(axis=0))
    return sym_groups

def get_symmetry_operations_to_generate_crystal_contacts(ref_hierarchy, crystal_symmetry, buffer_thickness):
    """Use an alternate method to identify the symmetry operations required to generate crystal contacts"""

    # Extract the xray structure from the reference hierarchy
    ref_struc = ref_hierarchy.extract_xray_structure(crystal_symmetry=crystal_symmetry)
    ref_atoms = ref_hierarchy.atoms()

    # Extract the mappings that will tell us the adjacent symmetry copies
    asu_mappings = ref_struc.asu_mappings(buffer_thickness=buffer_thickness)
    uc = asu_mappings.unit_cell()
    # Symmetry operations for each atom
    mappings = asu_mappings.mappings()

    # There should be one mappings list per atom
    assert len(ref_struc.scatterers()) == len(mappings)

    # Get all atom pairs within distance_cutoff distance
    pair_generator = crystal.neighbors_fast_pair_generator(
        asu_mappings,
        distance_cutoff=buffer_thickness)
    sym_operations = []
    for pair in pair_generator:
      # obtain rt_mx_ji - symmetry operator that should be applied to j-th atom
      # to transfer it to i-th atom
      rt_mx_i = asu_mappings.get_rt_mx_i(pair)
      rt_mx_j = asu_mappings.get_rt_mx_j(pair)
      rt_mx_ji = rt_mx_i.inverse().multiply(rt_mx_j)
      # if it is not a unit matrix, that is symmetry related pair of atoms
      if not rt_mx_ji.is_unit_mx():
        if rt_mx_ji not in sym_operations:
          sym_operations.append(rt_mx_ji)

    return sym_operations

def generate_crystal_copies_from_operations(ref_hierarchy, crystal_symmetry, sym_ops_mat):
    """Take a list of symmetry operations and apply them to reference hierarchy"""

    # Extract the xray structure from the reference hierarchy
    ref_struc = ref_hierarchy.extract_xray_structure(crystal_symmetry=crystal_symmetry)

    # Create a list of chain ids for the new symmetry copies
    # Already existing chain ids
    ref_chains = [c.id for c in ref_hierarchy.chains()]
    # Number of new chains needed - one for each reference chain for each symmetry operation
    num_new_chains = len(ref_chains)*len(sym_ops_mat)
    # Select the new chains, ignoring the ones already in the reference
    new_chain_ids = [c for c in iotbx.pdb.systematic_chain_ids()[0:(num_new_chains+len(ref_chains))] if c not in ref_chains][0:num_new_chains]

    assert not [c for c in new_chain_ids if c in ref_chains], 'GENERATED CHAIN IDS ARE NOT UNIQUE'

    # Create combinations of the different symmetry operations and the reference chains and map them to new chain ids
    sym_op_ref_chain_combinations = []
    for sym_op in sym_ops_mat:
        for r_chain in ref_chains:
            sym_op_ref_chain_combinations.append((sym_op.as_xyz(), r_chain))
    # Dictionary to map symmetry operations and reference chain ids to new chain ids
    new_chain_ids_hash = dict(zip(sym_op_ref_chain_combinations, new_chain_ids))

    # Hierarchies to be returned
    sym_hierarchies = []
    chain_mappings = dict([(c, []) for c in ref_chains])

    for sym_op in sym_ops_mat:

        sym_op_rt = sym_op.as_rational().as_float()

        # Transform all of the coordinates in the reference structure
        transformed_coords = sym_op_rt * ref_struc.sites_frac()

        # Create copy of the xray structure to play with, and set the transformed coordinates
        new_struc = ref_struc.customized_copy()
        new_struc.set_sites_frac(transformed_coords)

        # Transfer the sites to a new hierarchy
        new_hierarchy = ref_hierarchy.deep_copy()
        new_hierarchy.adopt_xray_structure(new_struc)

        # Update the chain ids in the new structure
        for chain in new_hierarchy.chains():
            old_id = chain.id
            chain.id = new_chain_ids_hash[(sym_op.as_xyz(), chain.id)]
            chain_mappings[old_id].append(chain.id)

        # Add the hierarchy to the output dict, referenced by the symmetry operation
        sym_hierarchies.append(new_hierarchy)

    return sym_hierarchies, chain_mappings

def generate_adjacent_symmetry_copies(ref_hierarchy, crystal_symmetry, buffer_thickness=0, method=2):
    """Find symmetry copies of the protein in contact with the asu and generate these copies"""

    sym_ops_mat = get_symmetry_operations_to_generate_crystal_contacts( ref_hierarchy=ref_hierarchy,
                                                                        crystal_symmetry=crystal_symmetry,
                                                                        buffer_thickness=buffer_thickness   )

    sym_hierarchies, chain_mappings = generate_crystal_copies_from_operations(ref_hierarchy=ref_hierarchy,
                                                                              crystal_symmetry=crystal_symmetry,
                                                                              sym_ops_mat=sym_ops_mat)

    return sym_ops_mat, sym_hierarchies, chain_mappings

if __name__=='__main__':

    input_file = './reference.pdb'

    inp = iotbx.pdb.input(input_file)
    hie = inp.construct_hierarchy()

    for method in [1,2]:
        sym_ops, contact_mappings, sym_hierarchies, chain_mappings = generate_adjacent_symmetry_copies(    ref_hierarchy=hie,
                                                                                                           crystal_symmetry=inp.crystal_symmetry(),
                                                                                                           buffer_thickness=50,
                                                                                                           method=method)

        print 'CHAIN MAPPINGS:'
        for ch in chain_mappings.keys():
            print '\tCHAIN {!s} maps to {!s}'.format(ch, chain_mappings[ch])
        print 'SYMMETRY HEIRARCHIES:'
        for x in sym_hierarchies:
            print '\t',x
        print 'SYMMETRY OPERATIONS:'
        for x in sorted(sym_ops, key=lambda m: str(m)):
            print '\t',x
        print '{!s} SYMMETRY COPIES GENERATED'.format(len(sym_hierarchies))

        combined_sym_hierarchy = combine_hierarchies(sym_hierarchies)

        output_file = input_file.replace('.pdb', '-contacts-method{!s}.pdb'.format(method))

        assert not os.path.exists(output_file)
        combined_sym_hierarchy.write_pdb_file(output_file)

